"""
This module provides functionality for resolving references within an instance of `oapi.oas.model.OpenAPI`.

For example, the following will replace all references in the Open API document `open_api_document` with the objects
targeted by the `ref` property of the reference:

```python
from urllib.request import urlopen
from oapi.oas.model import OpenAPI

with urlopen(
    'https://raw.githubusercontent.com/OAI/OpenAPI-Specification/master/'
    'examples/v3.0/callback-example.yaml'
) as response:
    open_api_document = OpenAPI(response)

resolver = Resolver(open_api_document)
resolver.dereference()
```
"""
import collections
from collections import OrderedDict
from itertools import chain
from typing import Dict, Optional, Sequence, Union
from urllib import request
from urllib.error import HTTPError
from urllib.parse import urljoin, urlparse
from warnings import warn

from jsonpointer import resolve_pointer

from sob import meta
from sob.errors import get_exception_text
from sob.model import (
    Array, Dictionary, Model, Object, detect_format, unmarshal
)
from sob.properties import Property
from sob.utilities.types import NoneType
from sob.utilities import qualified_name
from .model import OpenAPI, Reference
from ..errors import ReferenceLoopError


class _Document:

    def __init__(
        self,
        resolver,  # type: Resolver
        root,  # type: Union[Sequence, Dict, Array, Dictionary]
        url=None,  # type: Optional[str]
    ):
        # type: (...) -> None

        assert isinstance(url, str)
        assert isinstance(resolver, Resolver)

        # Infer the document URL
        if (url is None) and isinstance(root, Model):
            url = meta.url(root)

        self.resolver = resolver
        self.root = root
        self.pointers = {}
        self.url = url

    def get_url_pointer(self, pointer):
        # type: (str) -> Tuple[str, str]
        """
        Get an absolute URL + relative pointer
        """
        pointer_split = pointer.split('#')
        url = self.get_absolute_url(pointer_split[0])

        if len(pointer_split) > 1:
            pointer = '#'.join(pointer_split[1:])
            if pointer[0] != '#':
                pointer = '#' + pointer
        else:
            pointer = None

        return url, pointer

    def get_absolute_url(self, url):
        # type: (str) -> str
        """
        Get an absolute URL from a (possibly relative) URL
        """
        parse_result = urlparse(url)

        if not parse_result.scheme:
            url = urljoin(
                self.url,
                url.lstrip('/ ')
            )

        return url

    def dereference(self, model_instance, recursive=True):
        # type: (Model, bool) -> None
        """
        Recursively dereference this objects and all items/properties
        """
        try:
            if isinstance(model_instance, Object):
                self.dereference_object_properties(
                    model_instance, recursive=recursive
                )
            elif isinstance(
                model_instance,
                (Array, collections.Sequence, collections.Set)
            ) and not isinstance(
                model_instance,
                (str, bytes)
            ):
                self.dereference_array_items(
                    model_instance, recursive=recursive
                )
            elif isinstance(
                model_instance, (Dictionary, dict, OrderedDict)
            ):
                self.dereference_dictionary_values(
                    model_instance, recursive=recursive
                )
            else:
                raise TypeError(
                    'The argument must be an instance of `%s`, not %s' % (
                        qualified_name(Model),
                        repr(model_instance)
                    )
                )
        except ReferenceLoopError:
            if not recursive:
                raise

    def prevent_infinite_recursion(self, model_instance):
        # type: (Model) -> Model
        """
        Prevent recursion errors by putting a placeholder `None` in place of
        the parent object in the `pointer` cache
        """
        pointer = meta.pointer(model_instance)
        existing_value = None  # type: Model
        if pointer:
            if pointer in self.pointers:
                existing_value = self.pointers[pointer]
            self.pointers[pointer] = None
        return pointer, existing_value

    def reset_recursion_placeholder(self, pointer, previous_value):
        # type: (Optional[str], Optional[Model]) -> None
        """
        Cleanup a placeholder created by the `prevent_infinite_recursion`
        method
        """
        if pointer and (pointer in self.pointers):
            if previous_value is None:
                del self.pointers[pointer]
            else:
                self.pointers[pointer] = previous_value

    def dereference_object_properties(self, object_, recursive=True):
        # type: (Model, bool) -> None
        """
        Replace all references in this object's properties with the referenced
        object
        """
        object_meta = meta.read(object_)

        # Prevent recursion errors
        pointer, existing = self.prevent_infinite_recursion(object_)

        for property_name, property_ in object_meta.properties.items():
            value = getattr(object_, property_name)
            if isinstance(value, Reference):
                setattr(
                    object_,
                    property_name,
                    self.resolve(
                        value.ref,
                        types=(property_,),
                        dereference=recursive
                    )
                )
            elif recursive and isinstance(value, Model):
                self.dereference(value, recursive=recursive)

        self.reset_recursion_placeholder(pointer, existing)

    def dereference_array_items(self, array, recursive=True):
        # type: (Model, bool) -> None
        """
        Replace all references in this array with the referenced object
        """
        array_meta = meta.read(array)

        # Prevent recursion errors
        pointer, existing = self.prevent_infinite_recursion(array)

        for index in range(len(array)):
            item = array[index]
            if isinstance(item, Reference):
                array[index] = self.resolve(
                    item.ref,
                    types=array_meta.item_types,
                    dereference=recursive
                )
            elif recursive and isinstance(item, Model):
                self.dereference(item, recursive=recursive)

        self.reset_recursion_placeholder(pointer, existing)

    def dereference_dictionary_values(self, dictionary, recursive=True):
        # type: (Model, bool) -> None
        """
        Replace all references in this dictionary with the referenced object
        """
        dictionary_meta = meta.read(dictionary)

        # Prevent recursion errors
        pointer, existing = self.prevent_infinite_recursion(dictionary)

        for key, value in dictionary.items():
            if isinstance(value, Reference):
                dictionary[key] = self.resolve(
                    value.ref,
                    types=dictionary_meta.value_types,
                    dereference=recursive
                )
            elif recursive and isinstance(value, Model):
                self.dereference(value, recursive=recursive)

        self.reset_recursion_placeholder(pointer, existing)

    @staticmethod
    def unmarshal_resolved_reference(resolved_reference, url, pointer, types=None):
        # type: (Model, Optional[str], str, Sequence[Property, type]) -> Model
        if types or (not isinstance(resolved_reference, Model)):
            resolved_reference = unmarshal(resolved_reference, types=types)
            # Re-assign the URL and pointer
            meta.url(resolved_reference, url)
            meta.pointer(resolved_reference, pointer)
        return resolved_reference

    def resolve(self, pointer, types=None, dereference=False):
        # type: (str, Sequence[Property, type], bool) -> Model
        """
        Return the object referenced by a pointer
        """
        if pointer in self.pointers:
            # This catches recursion errors
            if self.pointers[pointer] is None:
                raise ReferenceLoopError(pointer)
        else:
            self.pointers[pointer] = None
            if pointer[0] == '#':
                # Resolve a reference within the same Open API document
                resolved = resolve_pointer(self.root, pointer[1:])
                # Cast the resolved reference as one of the given types
                resolved = self.unmarshal_resolved_reference(
                    resolved,
                    self.url,
                    pointer,
                    types=types
                )
                if resolved is None:
                    raise RuntimeError()
            else:
                # Resolve a reference from another Open API document
                url, document_pointer = self.get_url_pointer(pointer)
                # Retrieve the document
                document = self.resolver.get_document(
                    urljoin(self.url, url.lstrip('/'))
                )
                # Resolve the pointer, if needed
                if document_pointer:
                    resolved = document.resolve(document_pointer, types)
                else:
                    resolved = document.root
                    # Cast the resolved reference as one of the given types
                    resolved = self.unmarshal_resolved_reference(
                        resolved, url, document_pointer, types=types
                    )
            # Recursively dereference
            if dereference and isinstance(resolved, Model):
                self.dereference(resolved, recursive=dereference)
            # Cache the resolved pointer
            self.pointers[pointer] = resolved
        return self.pointers[pointer]

    def dereference_all(self):
        # type: (OpenAPI) -> None
        self.dereference(self.root, recursive=True)


class Resolver:
    """
    This class should be used, with an instance of `oapi.oas.model.OpenAPI`, to
    resolve references.

    Parameters:

        - root (oapi.oas.model.OpenAPI): The OpenAPI document against which
          pointers will be resolved.

        - url (str): The URL or file path from where `root` was retrieved. The
          base URL for relative paths will be the directory above this URL.
          This will not typically be needed, as it can be inferred from most
          `Model` instances.

        - urlopen (collections.Callable): If provided, this should be a
          function taking one argument (a `str`), which can be used in lieu
          of `urllib.request.urlopen` to retrieve a document and return an
          instance of a sub-class of `IOBase` (such as
          `http.client.HTTPResponse`). This should be used if authentication is
          needed in order to retrieve external references in the document,
          or if local file paths will be referenced instead of web URL's (use
          `open` as the value for the `urlopen` parameter
          in this case).
    """

    def __init__(self, root, url=None, urlopen = request.urlopen):
        # type: (OpenAPI, Optional[str], typing.Callable, bool) -> None
        # Ensure arguments are of the correct types
        assert callable(urlopen)
        assert isinstance(root, OpenAPI)
        assert isinstance(url, (str, NoneType))
        # This is the function used to open external pointer references
        self.urlopen = urlopen
        # Infer the URL from the `OpenAPI` document, if not explicitly provided
        if url is None:
            url = meta.url(root) or '#'
        self.url = url
        # This is the primary document--the one we are resolving
        document = _Document(self, root, url)
        # Store the primary document both by URL and under the key "#" (for
        # convenient reference)
        self.documents = {
            url: document
        }
        if url != '#':
            self.documents['#'] = document

    def get_document(self, url):
        # type: (str) -> _Document
        """
        Retrieve a document by URL, or use the cached document if previously
        retrieved
        """
        if url not in self.documents:
            try:
                with self.urlopen(url) as response:
                    self.documents[url] = _Document(
                        self, detect_format(response)[0], url=url
                    )
            except HTTPError as e:
                e.msg = '%s: %s' % (
                    get_exception_text().rstrip(),
                    url
                )
                raise e
            except FileNotFoundError as e:
                e.args = tuple(
                    chain(
                        [
                            '%s: %s' % (
                                e.args[0],
                                url
                            )
                        ],
                        e.args[1:]
                    ) if e.args else (
                        url
                    )
                )
                raise e
        return self.documents[url]

    def dereference(self):
        # type: (...) -> None
        """
        Dereference the primary document
        """
        self.documents['#'].dereference_all()

    def resolve(self, pointer, types=None, dereference=False):
        # type: (str, Sequence[Property, type], bool) -> Model
        url, pointer = self.documents['#'].get_url_pointer(pointer)
        return self.documents[url].resolve(pointer, types, dereference=dereference)



"""
Version 2x: https://swagger.io/docs/specification/2-0/basic-structure/
Version 3x: https://swagger.io/specification
"""

from copy import deepcopy
from numbers import Number
import typing
from typing import Union, Any

import sob


class Object(sob.model.Object):

    def __init__(
        self,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)


_object_hooks = sob.hooks.writable(Object)  # type: hooks.Object


def _add_property(self, key):
    # type: (Object, str) -> None
    """
    Look for a matching property, and if none exists--create one
    """
    try:
        self._get_key_property_name(key)
    except KeyError:
        meta_ = sob.meta.writable(self)
        if meta_.properties is None:
            meta_.properties = sob.meta.Properties()
        property_name = sob.utilities.property_name(key)
        meta_.properties[property_name] = sob.properties.Property(name=key)


def _object_before_setitem(self, key, value):
    # type: (Object, str, Any) -> Tuple[str, Any]
    """
    This hook allows for the use of extension attributes
    """
    if key[:2] == 'x-':
        _add_property(self, key)
    return key, value


_object_hooks.before_setitem = _object_before_setitem


class Array(sob.model.Array):

    pass


class Dictionary(sob.model.Dictionary):

    pass


class Reference(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        ref=None,  # type: Optional[str]
        **kwargs
    ):
        self.ref = ref
        super().__init__(_)
        for key, value in kwargs.items():
            self[key] = value


sob.meta.writable(Reference).properties = [
    ('ref', sob.properties.String(name='$ref'))
]


_reference_hooks = sob.hooks.writable(Reference)  # type: hooks.Object


def _reference_before_setitem(self, key, value):
    # type: (Object, str, Any) -> Tuple[str, Any]
    """
    This hook allows for the use of any arbitrary attribute, as specified in
    the `patternProperties` for this object in the OpenAPI schema
    """
    if key != '$ref':
        _add_property(self, key)
    return key, value


_reference_hooks.before_setitem = _reference_before_setitem


def _reference_after_unmarshal(data):  # noqa
    # type: (Dict[str, Any]) -> None
    """
    This ensures all reference objects have a `$ref` attribute
    """
    if data['$ref'] is None:
        raise TypeError(
            'All instances of `%s` must have a `$ref` attribute' %
            sob.utilities.qualified_name(Reference)
        )
    return data


_reference_hooks.after_unmarshal = _reference_after_unmarshal


class Contact(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md
    #contactObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        name=None,  # type: Optional[str]
        url=None,  # type: Optional[str]
        email=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        self.name = name
        self.url = url
        self.email = email
        super().__init__(_)


sob.meta.writable(Contact).properties = [
    ('name', sob.properties.String()),
    ('url', sob.properties.String()),
    ('email', sob.properties.String()),
]


class License(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md
    #licenseObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        name=None,  # type: Optional[str]
        url=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        self.name = name
        self.url = url
        super().__init__(_)


sob.meta.writable(License).properties = [
    ('name', sob.properties.String(required=True)),
    ('url', sob.properties.String()),
]


class Info(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#infoObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        title=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        terms_of_service=None,  # type: Optional[str]
        contact=None,  # type: Optional[Contact]
        license_=None,  # type: Optional[License]
        version=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        self.title = title
        self.description = description
        self.terms_of_service = terms_of_service
        self.contact = contact
        self.license_ = license_
        self.version = version
        super().__init__(_)


sob.meta.writable(Info).properties = [
    ('title', sob.properties.String(required=True)),
    ('description', sob.properties.String()),
    ('terms_of_service', sob.properties.String(name='termsOfService')),
    ('contact', sob.properties.Property(types=(Contact,))),
    ('license_', sob.properties.Property(types=(License,), name='license')),
    ('version', sob.properties.String(required=True)),
]


class Tag(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#tagObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        name=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        external_docs=None,  # type: Optional[ExternalDocumentation]
    ):
        # type: (...) -> None
        self.name = name
        self.description = description
        self.external_docs = external_docs
        super().__init__(_)


# Metadata definitions for `Tag` postponed until after `ExternalDocumentation` has been defined


class Link(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        rel=None,  # type: Optional[str]
        href=None,  # type: Optional[str]
    ):
        self.rel = rel
        self.href = href
        super().__init__(_)


sob.meta.writable(Link).properties = [
    ('rel', sob.properties.String()),
    ('href', sob.properties.String()),
]


class Schema(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#schemaObject
    http://json-schema.org

    Properties:

        - title (str)

        - description (str)

        - multiple_of (Number): The numeric value this schema describes should be divisible by this number.

        - maximum (Number): The number this schema describes should be less than or equal to this value, or less than
          this value, depending on the value of `exclusive_maximum`.

        - exclusive_maximum (bool): If `True`, the numeric instance described by this schema must be *less than*
          `maximum`. If `False`, the numeric instance described by this schema can be less than or *equal to*
          `maximum`.

        - minimum (Number): The number this schema describes should be greater than or equal to this value, or greater
          than this value, depending on the value of `exclusive_minimum`.

        - exclusive_minimum (bool): If `True`, the numeric instance described by this schema must be *greater than*
          `minimum`. If `False`, the numeric instance described by this schema can be greater than or *equal to*
          `minimum`.

        - max_length (int): The number of characters in the string instance described by this schema must be less than,
          or equal to, the value of this property.

        - min_length (int): The number of characters in the string instance described by this schema must be greater
          than, or equal to, the value of this property.

        - pattern (str): The string instance described by this schema should match this regular expression (ECMA 262).

        - items (Reference|Schema|[Schema]):

            - If `items` is a sub-schema--each item in the array instance described by this schema should be valid as
              described by this sub-schema.

            - If `items` is a sequence of sub-schemas, the array instance described by this schema should be equal in
              length to this sequence, and each value should be valid as described by the sub-schema at the
              corresponding index within this sequence of sub-schemas.

        - max_items (int): The array instance described by this schema should contain no more than this number of
          items.

        - min_items (int): The array instance described by this schema should contain at least this number of
          items.

        - unique_items (bool): The array instance described by this schema should contain only unique items.

        - max_properties (int)

        - min_properties (int)

        - sob.properties ({str:Schema}): Any sob.properties of the object instance described by this schema which
          correspond to a name in this mapping should be valid as described by the sub-schema corresponding to that name.

        - additional_properties (bool|Schema):

            - If `additional_properties` is `True`--sob.properties may be present in the object described by
              this schema with names which do not match those in `sob.properties`.

            - If `additional_properties` is `False`--all sob.properties present in the object described by this schema
              must correspond to a property matched in either `sob.properties`.

        - enum ([Any]): The value/instance described by this schema should be among those in this sequence.

        - type_ (str|[str]): The value/instance described by this schema should be of the value_types indicated
          (if this is a string), or *one of* the value_types indicated (if this is a sequence).

            - "null"
            - "boolean"
            - "object"
            - "array"
            - "number"
            - "string"

        - format_ (str|[str]):

            - "date-time": A date and time in the format YYYY-MM-DDThh:mm:ss.sTZD (eg 1997-07-16T19:20:30.45+01:00),
              YYYY-MM-DDThh:mm:ssTZD (eg 1997-07-16T19:20:30+01:00), or YYYY-MM-DDThh:mmTZD (eg 1997-07-16T19:20+01:00).
            - "email"
            - "hostname"
            - "ipv4"
            - "ipv6"
            - "uri"
            - "uriref": A URI or a relative reference.

        - all_of ([Schema]): The value/instance described by the schema should *also* be valid as
          described by all sub-schemas in this sequence.

        - any_of ([Schema]): The value/instance described by the schema should *also* be valid as
          described in at least one of the sub-schemas in this sequence.

        - one_of ([Schema]): The value/instance described by the schema should *also* be valid as
          described in one (but *only* one) of the sub-schemas in this sequence.

        - is_not (Schema): The value/instance described by this schema should *not* be valid as described by this
          sub-schema.

        - definitions ({str:Schema}): A dictionary of sub-schemas, stored for the purpose of referencing
          these sub-schemas elsewhere in the schema.

        - required ([str]): A list of attributes which must be present on the object instance described by this
          schema.

        - default (Any): The value presumed if the value/instance described by this schema is absent.

        The following sob.properties are specific to OpenAPI (not part of the core `JSON Schema <http://json-schema.org>`):

        - nullable (bool): If `True`, the value/instance described by this schema may be a null value (`None`).

        - discriminator (Discriminator): Adds support for polymorphism.

        - read_only (bool): If `True`, the property described may be returned as part of a response, but should not
          be part of a request.

        - write_only (bool): If `True`, the property described may be sent as part of a request, but should not
          be returned as part of a response.

        - xml (XML): Provides additional information describing XML representation of the property described by this
          schema.

        - external_docs (ExternalDocumentation)

        - example (Any)

        - definitions (Any)

        - depracated (bool): If `True`, the property or instance described by this schema should be phased out, as
          if will no longer be supported in future versions.
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        title=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        multiple_of=None,  # type: Optional[Number]
        maximum=None,  # type: Optional[Number]
        exclusive_maximum=None,  # type: Optional[bool]
        minimum=None,  # type: Optional[Number]
        exclusive_minimum=None,  # type: Optional[bool]
        max_length=None,  # type: Optional[int]
        min_length=None,  # type: Optional[int]
        pattern=None,  # type: Optional[str]
        items=None,  # type: Optional[Reference, Schema, Sequence[Schema]]
        max_items=None,  # type: Optional[int]
        min_items=None,  # type: Optional[int]
        unique_items=None,  # type: Optional[bool]
        max_properties=None,  # type: Optional[int]
        min_properties=None,  # type: Optional[int]
        properties=None,  # type: Optional[typing.Mapping[str, Schema]]
        additional_properties=None,  # type: Optional[bool, Schema]
        enum=None,  # type: Optional[Sequence]
        type_=None,  # type: Optional[str, Sequence]
        format_=None,  # type: Optional[str, Sequence]
        all_of=None,  # type: Optional[Sequence[Schema]]
        any_of=None,  # type: Optional[Sequence[Schema]]
        one_of=None,  # type: Optional[Sequence[Schema]]
        is_not=None,  # type: Optional[Schema]
        definitions=None,  # type: Optional[typing.Mapping[Schema]]
        required=None,  # type: Optional[Sequence[str]]
        default=None,  # type: Optional[Any]
        discriminator=None,  # type: Optional[Union[Discriminator, str]]
        read_only=None,  # type: Optional[bool]
        write_only=None,  # type: Optional[bool]
        xml=None,  # type: Optional[XML]
        external_docs=None,  # type: Optional[ExternalDocumentation]
        example=None,  # type: Any
        deprecated=None,  # type: Optional[bool]
        links=None,  # type: Optional[Sequence[Link]]
        nullable=None,  # type: Optional[bool]
    ):
        self.title = title
        self.description = description
        self.multiple_of = multiple_of
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.max_length = max_length
        self.min_length = min_length
        self.pattern = pattern
        self.items = items
        self.max_items = max_items
        self.min_items = min_items
        self.unique_items = unique_items
        self.max_properties = max_properties
        self.min_properties = min_properties
        self.properties = properties
        self.additional_properties = additional_properties
        self.enum = enum
        self.type_ = type_
        self.format_ = format_
        self.all_of = all_of
        self.any_of = any_of
        self.one_of = one_of
        self.is_not = is_not
        self.definitions = definitions
        self.required = required
        self.default = default
        self.discriminator = discriminator
        self.read_only = read_only
        self.write_only = write_only
        self.xml = xml
        self.external_docs = external_docs
        self.example = example
        self.deprecated = deprecated
        self.links = links
        self.nullable = nullable
        super().__init__(_)


# ...definitions are postponed until dependencies are defined


class Example(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#exampleObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        summary=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        value=None,  # type: Any
        external_value=None,  # type: Optional[str]
    ):
        self.summary = summary
        self.description = description
        self.value = value
        self.external_value = external_value
        super().__init__(_)


sob.meta.writable(Example).properties = [
    ('summary', sob.properties.String(versions=('openapi>=3.0',))),
    ('description', sob.properties.String(versions=('openapi>=3.0',))),
    ('value', sob.properties.Property(versions=('openapi>=3.0',))),
    ('external_value', sob.properties.String(name='externalValue', versions=('openapi>=3.0',))),
]


class Encoding(Object):

    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#encodingObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        content_type=None,  # type: Optional[str]
        headers=None,  # type: Optional[typing.Mapping[str, Union[Header, Reference]]]
        style=None,  # type: Optional[str]
        explode=None,  # type: Optional[bool]
        allow_reserved=None,  # type: Optional[bool]
    ):
        self.content_type = content_type
        self.headers = headers
        self.style = style
        self.explode = explode
        self.allow_reserved = allow_reserved
        super().__init__(_)


# ...definitions for `Encoding` are delayed until after the `Headers` definition


class MediaType(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#mediaTypeObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        schema=None,  # type: Optional[Schema, Reference]
        example=None,  # type: Any
        examples=None,  # type: Union[typing.Mapping[str, Union[Example, Reference]]]
        encoding=None,  # type: Union[typing.Mapping[str, Union[Encoding, Reference]]]
    ):
        self.schema = schema
        self.example = example
        self.examples = examples
        self.encoding = encoding
        super().__init__(_)


sob.meta.writable(MediaType).properties = [
    ('schema', sob.properties.Property(types=(Reference, Schema), versions=('openapi>=3.0',))),
    ('example', sob.properties.Property(versions=('openapi>=3.0',))),
    ('examples', sob.properties.Dictionary(value_types=(Reference, Example), versions=('openapi>=3.0',))),
    ('encoding', sob.properties.Dictionary(value_types=(Reference, Encoding), versions=('openapi>=3.0',))),
]


class Items(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        type_=None,  # type: Optional[str]
        format_=None,  # type: Optional[str, Sequence]
        items=None,  # type: Optional[Items]
        collection_format=None,  # type: Optional[str]
        default=None,  # type: Optional[Any]
        maximum=None,  # type: Optional[numbers.Number]
        exclusive_maximum=None,  # type: Optional[bool]
        minimum=None,  # type: Optional[numbers.Number]
        exclusive_minimum=None,  # type: Optional[bool]
        max_length=None,  # type: Optional[int]
        min_length=None,  # type: Optional[int]
        pattern=None,  # type: Optional[str]
        max_items=None,  # type: Optional[str]
        min_items=None,  # type: Optional[str]
        unique_items=None,  # type: Optional[bool]
        enum=None,  # type: Optional[Sequence[str]]
        multiple_of=None,  # type: Optional[Number]
    ):
        self.type_ = type_
        self.format_ = format_
        self.items = items
        self.collection_format = collection_format
        self.default = default
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.max_length = max_length
        self.min_length = min_length
        self.pattern = pattern
        self.max_items = max_items
        self.min_items = min_items
        self.unique_items = unique_items
        self.enum = enum
        self.multiple_of = multiple_of
        super().__init__(_)


sob.meta.writable(Items).properties = [
    (
        'type_',
        sob.properties.Enumerated(
            name='type',
            values=(
                'array',
                'object',
                'file',
                'integer',
                'number',
                'string',
                'boolean'
            ),
            versions=('openapi<3.0')
        )
    ),
    (
        'format_',
        sob.properties.String(
            # values=lambda o: (
            #     None
            #     if o is None else
            #     ('int32', 'int64')
            #     if o.type_ == 'integer' else
            #     ('float', 'double')
            #     if o.type_ == 'number' else
            #     ('byte', 'binary', 'date', 'date-time', 'password')
            #     if o.type_ == 'string'
            #     else tuple()
            # ),
            name='format',
            versions=('openapi<3.0')
        )
    ),
    (
        'items',
        sob.properties.Property(types=(Items,), versions=('openapi<3.0'))
    ),
    (
        'collection_format',
        sob.properties.Enumerated(
            values=('csv', 'ssv', 'tsv', 'pipes'),
            name='collectionFormat',
            versions=('openapi<3.0')
        )
    ),
    (
        'default',
        sob.properties.Property()
    ),(
        'maximum',
        sob.properties.Number(versions=('openapi<3.0',))
    ),
    (
        'exclusive_maximum',
        sob.properties.Boolean(
            name='exclusiveMaximum',
            versions=('openapi<3.0',)
        )
    ),
    (
        'minimum',
        sob.properties.Number(versions=('openapi<3.0',))
    ),
    (
        'exclusive_minimum',
        sob.properties.Boolean(
            name='exclusiveMinimum',
            versions=('openapi<3.0',)
        )
    ),
    (
        'max_length',
        sob.properties.Integer(name='maxLength', versions=('openapi<3.0',))
    ),
    (
        'min_length',
        sob.properties.Integer(name='minLength', versions=('openapi<3.0',))
    ),
    (
        'pattern',
        sob.properties.String(versions=('openapi<3.0',))
    ),
    (
        'max_items',
        sob.properties.Integer(name='maxItems', versions=('openapi<3.0',))
    ),
    (
        'min_items',
        sob.properties.Integer(name='minItems', versions=('openapi<3.0',))
    ),
    (
        'unique_items',
        sob.properties.Boolean(
            name='uniqueItems',
            versions=('openapi<3.0',)
        )
    ),
    (
        'enum',
        sob.properties.Array(versions=('openapi<3.0',))
    ),
    (
        'multiple_of',
        sob.properties.Number(
            name='multipleOf',
            versions=('openapi<3.0',)
        )
    )
]


class Parameter(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#parameterObject

    Properties:

        - name (str)

        - in_ (str):

            - "path"
            - "query"
            - "header"
            - "cookie"

        - description (str)

        - required (bool)

        - deprecated (bool)

        - allow_empty_value (bool): Sets the ability to pass empty-valued parameters. This is valid only for query
          parameters and allows sending a parameter with an empty value. The default value is `False`. If `style`
          is used, and if `behavior` is inapplicable (cannot be serialized), the value of `allow_empty_value` will
          be ignored.

        - style (str): Describes how the parameter value will be serialized, depending on the type of the parameter
          value.

            - "matrix": Path-style parameters defined by `RFC6570 <https://tools.ietf.org/html/rfc6570#section-3.2.7>`.
            - "label": Label-style parameters defined by `RFC6570 <https://tools.ietf.org/html/rfc6570#section-3.2.5>`.
            - "form": Form style parameters defined by `RFC6570 <https://tools.ietf.org/html/rfc6570#section-3.2.8>`.
            - "simple": Simple style parameters defined by
              `RFC6570 <https://tools.ietf.org/html/rfc6570#section-3.2.2>`.
            - "spaceDelimited": Space separated array values.
            - "pipeDelimited": Pipe separated array values.
            - "deepObject": Provides a simple way of rendering nested objects using form parameters.

          Default values (based on value of `location`):

            - query: "form"
            - path: "simple"
            - header: "simple"
            - cookie: "form"

          https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#style-values-52

        - explode (bool): When this is `True`, array or object parameter values generate separate parameters for
          each value of the array or name-value pair of the map. For other value_types of parameters this property has no
          effect. When `style` is "form", the default value is `True`. For all other styles, the default value is
          `False`.

        - allow_reserved (bool): Determines whether the parameter value SHOULD allow reserved characters
          :/?#[]@!$&'()*+,;= (as defined by `RFC3986 <https://tools.ietf.org/html/rfc3986#section-2.2>`) to be included
          without percent-encoding. This property only applies to parameters with a location value of "query". The
          default value is `False`.

        - schema (Schema): The schema defining the type used for the parameter.

        - example (Any): Example of the media type. The example should match the specified schema and encoding
          sob.properties if present. The `example` parameter should not be present if `examples` is present. If
          referencing a `schema` which contains an example--*this* example overrides the example provided by the
          `schema`. To represent examples of media value_types that cannot naturally be represented in JSON or YAML, a
          string value can contain the example with escaping where necessary.

        - examples ({str:Example}): Examples of the media type. Each example should contain a value in the correct
          format, as specified in the parameter encoding. The `examples` parameter should not be present if
          `example` is present. If referencing a `schema` which contains an example--*these* example override the
          example provided by the `schema`. To represent examples of media value_types that cannot naturally be represented
          in JSON or YAML, a string value can contain the example with escaping where necessary.

          https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#format

        - content ({str:MediaType}): A map containing the representations for the parameter. The name is the media type
          and the value describing it. The map must only contain one entry.

    ...for version 2x compatibility:

        - type_ (str)

        - enum ([Any])
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        name=None,  # type: Optional[str]
        in_=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        required=None,  # type: Optional[bool]
        deprecated=None,  # type: Optional[bool]
        allow_empty_value=None, # type: Optional[bool]
        style=None,  # type: Optional[str]
        explode=None, # type: Optional[bool]
        allow_reserved=None, # type: Optional[bool]
        schema=None, # type: Optional[Schema]
        example=None, # type: Any
        examples=None, # type: Optional[typing.Mapping[str, Example]]
        content=None,  # type: Optional[typing.Mapping[str, MediaType]]
        type_=None,  # type: Optional[str]
        default=None,  # type: Any
        maximum=None,  # type: Optional[numbers.Number]
        exclusive_maximum=None,  # type: Optional[bool]
        minimum=None,  # type: Optional[numbers.Number]
        exclusive_minimum=None,  # type: Optional[bool]
        max_length=None,  # type: Optional[int]
        min_length=None,  # type: Optional[int]
        pattern=None,  # type: Optional[str]
        max_items=None,  # type: Optional[str]
        min_items=None,  # type: Optional[str]
        unique_items=None,  # type: Optional[bool]
        format_=None,  # type: Optional[str]
        enum=None,  # type: Optional[Sequence[str]]
        multiple_of=None,  # type: Optional[Number]
        collection_format=None,  # type: Optional[str]
        items=None,  # type: Optional[Schema]
    ):
        self.name = name
        self.in_ = in_
        self.description = description
        self.required = required
        self.deprecated = deprecated
        self.allow_empty_value = allow_empty_value
        self.style = style
        self.explode = explode
        self.allow_reserved = allow_reserved
        self.schema = schema
        self.example = example
        self.examples = examples
        self.content = content
        self.type_ = type_
        self.default = default
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.max_length = max_length
        self.min_length = min_length
        self.pattern = pattern
        self.max_items = max_items
        self.min_items = min_items
        self.unique_items = unique_items
        self.enum = enum
        self.multiple_of = multiple_of
        self.format_ = format_
        self.collection_format = collection_format
        self.items = items
        super().__init__(_)


sob.meta.writable(Parameter).properties = [
    ('name', sob.properties.String(required=True)),
    (
        'in_',
        sob.properties.Enumerated(
            values=('query', 'header', 'path', 'formData', 'body'),
            name='in',
            required=True
        )
    ),
    ('description', sob.properties.String()),
    ('required', sob.properties.Boolean()),
    ('deprecated', sob.properties.Boolean()),
    ('allow_empty_value', sob.properties.Boolean(name='allowEmptyValue')),
    ('style', sob.properties.String(versions=('openapi>=3.0',))),
    ('explode', sob.properties.Boolean(versions=('openapi>=3.0',))),
    ('allow_reserved', sob.properties.Boolean(name='allowReserved', versions=('openapi>=3.0',))),
    (
        'schema',
        sob.properties.Property(
            types=(Reference, Schema),
        )
    ),
    ('example', sob.properties.Property(versions=('openapi>=3.0',))),
    (
        'examples',
        sob.properties.Dictionary(
            value_types=(Reference, Example),
            versions=('openapi>=3.0',)
        )
    ),
    (
        'content',
        sob.properties.Dictionary(
            value_types=(MediaType,),
            versions=('openapi>=3.0')
        ),
    ),
    (
        'type_',
        sob.properties.Enumerated(
            values=(
                'array',
                'object',
                'file',
                'integer',
                'number',
                'string',
                'boolean'
            ),
            name='type',
            versions=('openapi<3.0',)
        )
    ),
    (
        'default',
        sob.properties.Property(
            versions=('openapi<3.0',)
        )
    ),
    (
        'maximum',
        sob.properties.Number(versions=('openapi<3.0',))
    ),
    (
        'exclusive_maximum',
        sob.properties.Boolean(
            name='exclusiveMaximum',
            versions=('openapi<3.0',)
        )
    ),
    (
        'minimum',
        sob.properties.Number(versions=('openapi<3.0',))
    ),
    (
        'exclusive_minimum',
        sob.properties.Boolean(
            name='exclusiveMinimum',
            versions=('openapi<3.0',)
        )
    ),
    (
        'max_length',
        sob.properties.Integer(name='maxLength', versions=('openapi<3.0',))
    ),
    (
        'min_length',
        sob.properties.Integer(name='minLength', versions=('openapi<3.0',))
    ),
    (
        'pattern',
        sob.properties.String(versions=('openapi<3.0',))
    ),
    (
        'max_items',
        sob.properties.Integer(name='maxItems', versions=('openapi<3.0',))
    ),
    (
        'min_items',
        sob.properties.Integer(name='minItems', versions=('openapi<3.0',))
    ),
    (
        'unique_items',
        sob.properties.Boolean(
            name='uniqueItems',
            versions=('openapi<3.0',)
        )
    ),
    (
        'enum',
        sob.properties.Array(versions=('openapi<3.0',))
    ),
    (
        'format_',
        sob.properties.String(
            # values=(
            #     'int32', 'int64',  # type_ == 'integer'
            #     'float', 'double',  # type_ == 'number'
            #     'byte', 'binary', 'date', 'date-time', 'password',  # type_ == 'string'
            # ),
            name='format',
            versions=('openapi<3.0')
        )
    ),
    (
        'collection_format',
        sob.properties.Enumerated(
            values=('csv', 'ssv', 'tsv', 'pipes', 'multi'),
            name='collectionFormat',
            versions=('openapi<3.0',)
        )
    ),
    (
        'items',
        sob.properties.Property(
            types=(Items,),
            versions=('openapi<3.0',)
        )
    ),
    (
        'multiple_of',
        sob.properties.Number(
            name='multipleOf',
            versions=('openapi<3.0',)
        )
    )
]


def _parameter_after_validate(o):
    # type: (Parameter) -> Parameter
    if (o.content is not None) and len(tuple(o.content.keys())) > 1:
        raise sob.errors.ValidationError(
            '`%s.content` may have only one mapped value.:\n%s' % (
                sob.utilities.qualified_name(type(o)), repr(o)
            )
        )
    if (o.content is not None) and (o.schema is not None):
        raise sob.errors.ValidationError(
            'An instance of `%s` may have a `schema` property or a `content` ' % sob.utilities.qualified_name(type(o)) +
            'property, but not *both*:\n' + repr(o)
        )
    if o.format_ in (
        'int32', 'int64',  # type_ == 'integer'
        'float', 'double',  # type_ == 'number'
        'byte', 'binary', 'date', 'date-time', 'password',  # type_ == 'string'
    ):
        if o.type_ == 'integer' and (
            o.format_ not in ('int32', 'int64', None)
        ):
            qn = sob.utilities.qualified_name(type(o))
            raise sob.errors.ValidationError(
                '"%s" in not a valid value for `%S.format_` in this circumstance. ' % (
                    o.format_,
                    qn
                ) +
                '`%s.format_` may be "int32" or "int64" when ' % qn +
                '`%s.type_` is "integer".' % qn
            )
        elif o.type_ == 'number' and (
            o.format_ not in ('float', 'double', None)
        ):
            qn = sob.utilities.qualified_name(type(o))
            raise sob.errors.ValidationError(
                '"%s" in not a valid value for `%s.format_` in this circumstance. ' % (o.format_, qn) +
                '`%s.format_` may be "float" or "double" when ' % qn +
                '`%s.type_` is "number".' % (qn, )
            )
        elif o.type_ == 'string' and (
            o.format_ not in ('byte', 'binary', 'date', 'date-time', 'password', None)
        ):
            qn = sob.utilities.qualified_name(type(o))
            raise sob.errors.ValidationError(
                '"%s" in not a valid value for `%s.format_` in this circumstance. ' % (o.format_, qn) +
                '`%s.format_` may be "byte", "binary", "date", "date-time" or "password" when ' % qn +
                '`%s.type_` is "string".' % (qn, )
            )
    return o


sob.hooks.writable(Parameter).after_validate = _parameter_after_validate


class Header(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        description=None,  # type: Optional[str]
        required=None,  # type: Optional[bool]
        deprecated=None,  # type: Optional[bool]
        allow_empty_value=None, # type: Optional[bool]
        style=None,  # type: Optional[str]
        explode=None, # type: Optional[bool]
        allow_reserved=None, # type: Optional[bool]
        schema=None, # type: Optional[Schema]
        example=None, # type: Any
        examples=None, # type: Optional[typing.Mapping[str, Example]]
        content=None,  # type: Optional[typing.Mapping[str, MediaType]]
        type_=None,  # type: Optional[str]
        default=None,  # type: Any
        maximum=None,  # type: Optional[numbers.Number]
        exclusive_maximum=None,  # type: Optional[bool]
        minimum=None,  # type: Optional[numbers.Number]
        exclusive_minimum=None,  # type: Optional[bool]
        max_length=None,  # type: Optional[int]
        min_length=None,  # type: Optional[int]
        pattern=None,  # type: Optional[str]
        max_items=None,  # type: Optional[str]
        min_items=None,  # type: Optional[str]
        unique_items=None,  # type: Optional[bool]
        format_=None,  # type: Optional[str]
        enum=None,  # type: Optional[Sequence[str]]
        multiple_of=None,  # type: Optional[Number]
        collection_format=None,  # type: Optional[str]
        items=None,  # type: Optional[Schema]
    ):
        self.description = description
        self.required = required
        self.deprecated = deprecated
        self.allow_empty_value = allow_empty_value
        self.style = style
        self.explode = explode
        self.allow_reserved = allow_reserved
        self.schema = schema
        self.example = example
        self.examples = examples
        self.content = content
        self.type_ = type_
        self.default = default
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.max_length = max_length
        self.min_length = min_length
        self.pattern = pattern
        self.max_items = max_items
        self.min_items = min_items
        self.unique_items = unique_items
        self.enum = enum
        self.multiple_of = multiple_of
        self.format_ = format_
        self.collection_format = collection_format
        self.items = items
        super().__init__(_)

_header_meta = sob.meta.writable(Header)
_header_meta.properties = deepcopy(sob.meta.read(Parameter).properties)
del _header_meta.properties['name']
del _header_meta.properties['in_']
_header_meta.properties['schema'].versions = ('openapi>=3.0',)
_header_meta.properties['type_'].values = (
    'array',
    'integer',
    'number',
    'string',
    'boolean'
)
_header_meta.properties['items'].types = (Items,)
_header_meta.properties['items'].required = lambda o: True if o.type_ == 'array' else False
_header_meta.properties['required'].versions = ('openapi>=3.0',)
_header_meta.properties['deprecated'].versions = ('openapi>=3.0',)
_header_meta.properties['allow_empty_value'].versions = ('openapi>=3.0',)


sob.meta.writable(Encoding).properties = [
    ('content_type', sob.properties.String(name='contentType')),
    ('headers', sob.properties.Dictionary(value_types=(Reference, Header))),
    ('style', sob.properties.String()),
    ('explode', sob.properties.Boolean()),
    ('allow_reserved', sob.properties.Boolean(name='allowReserved')),
]


class ServerVariable(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        enum=None,  # type: Optional[Sequence[str]]
        default=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
    ):
        self.enum = enum
        self.default = default
        self.description = description
        super().__init__(_)


sob.meta.writable(ServerVariable).properties = [
    ('enum', sob.properties.Array(item_types=(str,))),
    ('default', sob.properties.String(required=True)),
    ('description', sob.properties.String()),
]


class Server(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#serverObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        url=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        variables=None  # type: Optional[typing.Mapping[str, ServerVariable]]
    ):
        # type: (...) -> None
        self.url = url
        self.description = description
        self.variables = variables
        super().__init__(_)


sob.meta.writable(Server).properties = [
    ('url', sob.properties.String(required=True)),
    ('description', sob.properties.String()),
    ('variables', sob.properties.Dictionary(value_types=(ServerVariable,))),
]


class Link_(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#linkObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        operation_ref=None,  # type: Optional[str]
        operation_id=None,  # type: Optional[str]
        parameters=None,  # type: Optional[typing.Mapping[str, Any]]
        request_body=None,  # type: Any
        description=None,  # type: Optional[str]
        server=None,  # type: Optional[Server]
    ):
        self.operation_ref = operation_ref
        self.operation_id = operation_id
        self.parameters = parameters
        self.request_body = request_body
        self.description = description
        self.server = server
        super().__init__(_)


sob.meta.writable(Link_).properties = [
    ('operation_ref', sob.properties.String(name='operationRef', versions=('openapi>=3.0',))),
    ('operation_id', sob.properties.String(name='operationId', versions=('openapi>=3.0',))),
    ('parameters', sob.properties.Dictionary(versions=('openapi>=3.0',))),
    ('request_body', sob.properties.Property(name='requestBody', versions=('openapi>=3.0',))),
    ('description', sob.properties.String(versions=('openapi>=3.0',))),
    ('server', sob.properties.Property(types=(Server,), versions=('openapi>=3.0',))),
]


class Response(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#responseObject

    Properties:

        - description (str): A short description of the response. `CommonMark syntax<http://spec.commonmark.org/>` may
          be used for rich text representation.

        - headers ({str:Header|Reference}): Maps a header name to its definition (mappings are case-insensitive).

        - content ({str:Content|Reference}): A mapping of media value_types to `MediaType` instances describing potential
          payloads.

        - links ({str:Link_|Reference}): A map of operations links that can be followed from the response.
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        description=None,  # type: Optional[str]
        schema=None,  # type: Optional[Schema]
        headers=None,  # type: Optional[typing.Mapping[str, Union[Header, Reference]]]
        examples=None,  # type: Optional[Dict[str, Any]]
        # example=None,  # type: Optional[Any]
        content=None,  # type: Optional[typing.Mapping[str, Union[Content, Reference]]]
        links=None,  # type: Optional[typing.Mapping[str, Union[Link, Reference]]]
    ):
        # type: (...) -> None
        self.description = description
        self.schema = schema
        self.headers = headers
        self.examples = examples
        # self.example = example
        self.content = content
        self.links = links
        super().__init__(_)


sob.meta.writable(Response).properties = [
    (
        'description',
        sob.properties.String()
    ),
    (
        'schema',
        sob.properties.Property(
            types=(Reference, Schema),
            versions=('openapi<3.0',)
        )
    ),
    (
        'headers',
        sob.properties.Dictionary(
            value_types=(Reference, Header)
        )
    ),
    (
        'examples',
        sob.properties.Dictionary(
            versions=('openapi<3.0',)
        )
    ),
    (
        'content',
        sob.properties.Dictionary(
            value_types=(Reference, MediaType),
            versions=('openapi>=3.0',)
        )
    ),
    (
        'links',
        sob.properties.Dictionary(
            value_types=(Reference, Link_),
            versions=('openapi>=3.0',)
        )
    ),
]


class ExternalDocumentation(Object):
    """
    Properties:

        - description (str)
        - url (str)
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        description=None,  # type: Optional[str]
        url=None,  # type: Optional[str]
    ):
        self.description = description
        self.url = url
        super().__init__(_)


sob.meta.writable(ExternalDocumentation).properties = [
    ('description', sob.properties.String()),
    ('url', sob.properties.String(required=True)),
]


sob.meta.writable(Tag).properties = [
    ('name', sob.properties.String(required=True)),
    ('description', sob.properties.String()),
    ('external_docs', sob.properties.Property(types=(ExternalDocumentation,))),
]


class RequestBody(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        description=None,  # type: Optional[str]
        content=None,  # type: Optional[typing.Mapping[str, MediaType]]
        required=None,  # type: Optional[bool]
    ):
        self.description = description
        self.content = content
        self.required = required
        super().__init__(_)


sob.meta.writable(RequestBody).properties = [
    ('description', sob.properties.String(versions=('openapi>=3.0',))),
    ('content', sob.properties.Dictionary(value_types=(MediaType,), versions=('openapi>=3.0',))),
    ('required', sob.properties.Boolean(versions=('openapi>=3.0',))),
]


class Operation(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#operationObject

    Describes a single API operation on a path.

    Properties:

        - tags ([str]):  A list of tags for API documentation control. Tags can be used for logical grouping of
          operations by resources or any other qualifier.

        - summary (str):  A short summary of what the operation does.

        - description (str): A verbose explanation of the operation behavior. `CommonMark <http://spec.commonmark.org>`
          syntax may be used for rich text representation.

        - external_docs (ExternalDocumentation):  Additional external documentation for this operation.

        - operation_id (str):  Unique string used to identify the operation. The ID must be unique among all operations
          described in the API. Tools and libraries may use the `operation_id` to uniquely identify an operation,
          therefore, it is recommended to follow common programming naming conventions.

        - parameters ([Parameter|Reference]):  A list of parameters that are applicable for this operation. If a
          parameter is already defined at the `PathItem`, the new definition will override it, but can never remove
          it.

        - request_body (RequestBody|Reference):  The request body applicable for this operation. The requestBody is only
          supported in HTTP methods where the HTTP 1.1 specification
          `RFC7231 <https://tools.ietf.org/html/rfc7231#section-4.3.1>` has explicitly defined semantics for request
          bodies.

        - responses (typing.Mapping[str, Response]):  A mapping of HTTP response codes to response schemas.

        - callbacks ({str:{str:PathItem}|Reference})

        - deprecated (bool)

        - security ([[str]])

        - servers ([Server])

    Version 2x Compatibility:

        - produces ([str]):
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        tags=None,  # type: Optional[Sequence[str]]
        summary=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        external_docs=None,  # type: Optional[ExternalDocumentation]
        operation_id=None,  # type: Optional[str]
        parameters=None,  # type: Optional[Sequence[Union[Parameter, Reference]]]
        request_body=None,  # type: Optional[RequestBody, Reference]
        schemes=None,  # type: Optional[typing.Sequence[str]]
        deprecated=None,  # type: Optional[bool]
        responses=None,  # type: Optional[typing.Mapping[str, Response]]
        callbacks=None,  # type: Optional[typing.Mapping[str, Union[Callback, Refefence]]]
        security=None,  # type: Optional[Sequence[Dict[str, Sequence[str]]]]
        servers=None,  # type: Optional[Sequence[Server]]
        consumes=None,  # type: Optional[Sequence[str]]
        produces=None,  # type: Optional[Sequence[str]]
    ):
        # type: (...) -> None
        self.tags = tags
        self.summary = summary
        self.description = description
        self.external_docs = external_docs
        self.operation_id = operation_id
        self.parameters = parameters
        self.request_body = request_body
        self.responses = responses
        self.schemes = schemes
        self.deprecated = deprecated
        self.callbacks = callbacks
        self.security = security
        self.servers = servers
        self.consumes = consumes
        self.produces = produces
        super().__init__(_)


# Property definitions for `Operation` deferred until after `PathItem` is defined


class PathItem(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#pathItemObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        summary=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        get=None,  # type: Optional[Operation]
        put=None,  # type: Optional[Operation]
        post=None,  # type: Optional[Operation]
        delete=None,  # type: Optional[Operation]
        options=None,  # type: Optional[Operation]
        head=None,  # type: Optional[Operation]
        patch=None,  # type: Optional[Operation]
        trace=None,  # type: Optional[Operation]
        servers=None,  # type: Optional[Sequence[Server]]
        parameters=None,  # type: Optional[Sequence[Parameter]]
    ):
        # type: (...) -> None
        self.summary = summary
        self.description = description
        self.get = get
        self.put = put
        self.post = post
        self.delete = delete
        self.options = options
        self.head = head
        self.patch = patch
        self.trace = trace
        self.servers = servers
        self.parameters = parameters
        super().__init__(_)


sob.meta.writable(PathItem).properties = [
    ('summary', sob.properties.String(versions=('openapi>=3.0',))),
    ('description', sob.properties.String(versions=('openapi>=3.0',))),
    ('get', sob.properties.Property(types=(Operation,))),
    ('put', sob.properties.Property(types=(Operation,))),
    ('post', sob.properties.Property(types=(Operation,))),
    ('delete', sob.properties.Property(types=(Operation,))),
    ('options', sob.properties.Property(types=(Operation,))),
    ('head', sob.properties.Property(types=(Operation,))),
    ('patch', sob.properties.Property(types=(Operation,))),
    ('trace', sob.properties.Property(types=(Operation,), versions=('openapi>=3.0',))),
    ('servers', sob.properties.Array(item_types=(Server,), versions=('openapi>=3.0',))),
    (
        'parameters',
        sob.properties.Array(
            item_types=(Reference, Parameter)
        )
    ),
]


class Callback(Dictionary):

    pass


sob.meta.writable(Callback).value_types = (PathItem,)


class Callbacks(Dictionary):

    pass


sob.meta.writable(Callbacks).value_types = (Reference, Callback)


class Responses(Dictionary):

    pass


sob.meta.writable(Responses).value_types = (
    sob.properties.Property(
        types=(Reference, Response),
        versions=('openapi>=3.0',)
    ),
    sob.properties.Property(
        types=(Response,),
        versions=('openapi<3.0',)
    )
)


sob.meta.writable(Operation).properties = [
    ('tags', sob.properties.Array(item_types=(str,))),
    ('summary', sob.properties.String()),
    ('description', sob.properties.String()),
    ('external_docs', sob.properties.Property(types=(ExternalDocumentation,), name='externalDocs')),
    ('operation_id', sob.properties.String(name='operationId')),
    ('consumes', sob.properties.Array(item_types=(str,),versions=('openapi<3.0',))),
    ('produces', sob.properties.Array(item_types=(str,),versions=('openapi<3.0',))),
    (
        'parameters',
        sob.properties.Array(
            item_types=(Reference, Parameter)
        )
    ),
    (
        'request_body',
        sob.properties.Property(
            types=(Reference, RequestBody),
            name='requestBody',
            versions=('openapi>=3.0',)
        )
    ),
    (
        'responses',
        sob.properties.Property(
            types=(Responses,),
            required=True
        )
    ),
    (
        'callbacks',
        sob.properties.Property(
            types=(
                Callbacks,
            ),
            versions=('openapi>=3.0',)
        )
    ),
    ('schemes', sob.properties.Array(item_types=(str,), versions=('openapi>=3.0',))),
    ('deprecated', sob.properties.Boolean()),
    (
        'security',
        sob.properties.Array(
            item_types=(
                sob.properties.Dictionary(
                    value_types=(
                        sob.properties.Array(
                            item_types=(str,)
                        ),
                    )
                ),
            )
        )
    ),
    ('servers', sob.properties.Array(item_types=(Server,), versions=('openapi>=3.0',)))
]


class Discriminator(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#discriminatorObject

    Properties:

        - property_name (str): The name of the property which will hold the discriminating value.

        - mapping ({str:str}): An mappings of payload values to schema names or references.
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        property_name=None,  # type: Optional[str]
        mapping=None,  # type: Optional[typing.Mapping[str, str]]
    ):
        self.property_name = property_name
        self.mapping = mapping
        super().__init__(_)


sob.meta.writable(Discriminator).properties = [
    ('property_name', sob.properties.String(name='propertyName', versions=('openapi>=3.0',))),
    ('mapping', sob.properties.Dictionary(value_types=(str,), versions=('openapi>=3.0',))),
]


class XML(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#xmlObject

    Properties:

        - name (str): The element name.

        - name_space (str): The *absolute* URI of a namespace.

        - prefix (str): The prefix to be used with the name to reference the name-space.

        - attribute (bool): If `True`, the property described is an attribute rather than a sub-element.

        - wrapped (bool): If `True`, an array instance described by the schema will be wrapped by a tag (named
          according to the parent element's property, while `name` refers to the child element name).
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        name=None,  # type: Optional[str]
        name_space=None,  # type: Optional[str]
        prefix=None,  # type: Optional[str]
        attribute=None,  # type: Optional[bool]
        wrapped=None,  # type: Optional[bool]
    ):
        self.name = name
        self.name_space = name_space
        self.prefix = prefix
        self.attribute = attribute
        self.wrapped = wrapped
        super().__init__(_)


sob.meta.writable(XML).properties = [
    ('name', sob.properties.String()),
    ('name_space', sob.properties.String(name='nameSpace')),
    ('prefix', sob.properties.String()),
    ('attribute', sob.properties.Boolean()),
    ('wrapped', sob.properties.Boolean()),
]


class OAuthFlow(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#oauthFlowObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        authorization_url=None,  # type: Optional[str]
        token_url=None,  # type: Optional[str]
        refresh_url=None,  # type: Optional[str]
        scopes=None,  # type: Optional[typing.Mapping[str, str]]
    ):
        self.authorization_url = authorization_url
        self.token_url = token_url
        self.refresh_url = refresh_url
        self.scopes = scopes
        super().__init__(_)


sob.meta.writable(OAuthFlow).properties = [
    ('authorization_url', sob.properties.String()),
    ('token_url', sob.properties.String(name='tokenUrl')),
    ('refresh_url', sob.properties.String(name='refreshUrl')),
    ('scopes', sob.properties.Dictionary(value_types=(str,))),
]


class OAuthFlows(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#oauthFlowsObject
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        implicit=None,  # type: Optional[OAuthFlow]
        password=None,  # type: Optional[OAuthFlow]
        client_credentials=None,  # type: Optional[OAuthFlow]
        authorization_code=None,  # type: Optional[OAuthFlow]
    ):
        self.implicit = implicit
        self.password = password
        self.client_credentials = client_credentials
        self.authorization_code = authorization_code
        super().__init__(_)


sob.meta.writable(OAuthFlows).properties = [
    ('implicit', sob.properties.Property(types=(OAuthFlow,), versions=('openapi>=3.0',))),
    ('password', sob.properties.Property(types=(OAuthFlow,), versions=('openapi>=3.0',))),
    (
        'client_credentials',
        sob.properties.Property(
            types=(OAuthFlow,),
            name='clientCredentials',
            versions=('openapi>=3.0',),
        )
    ),
    (
        'authorization_code',
        sob.properties.Property(
            types=(OAuthFlow,),
            name='authorizationCode',
            versions=('openapi>=3.0',)
        )
    ),
]


class Properties(Dictionary):

    pass


sob.meta.writable(Properties).value_types = (Reference, Schema)


class SecurityScheme(Object):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#requestBodyObject

    Properties:

        - type_ (str): https://tools.ietf.org/html/rfc7235#section-4

        - description (str)

        - name (str)

        - in_ (str)

        - scheme (str)

        - bearer_format (str)

        - flows (OAuthFlows)

        - open_id_connect_url (str)
    """

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        type_=None,  # type: Optional[str]
        description=None,  # type: Optional[str]
        name=None,  # type: Optional[str]
        in_=None,  # type: Optional[str]
        scheme=None,  # type: Optional[str]
        bearer_format=None,  # type: Optional[str]
        flows=None,  # type: Optional[OAuthFlows]
        open_id_connect_url=None,  # type: Optional[str]
        flow=None,  # type: Optional[str]
        authorization_url=None,  # type: Optional[str]
        scopes=None,  # type: Optional[str]
        token_url=None,  # type: Optional[str]
    ):
        self.type_ = type_
        self.description = description
        self.name = name
        self.in_ = in_
        self.scheme = scheme
        self.bearer_format = bearer_format
        self.flows = flows
        self.open_id_connect_url = open_id_connect_url
        self.flow = flow
        self.authorization_url = authorization_url
        self.scopes = scopes
        self.token_url = token_url
        super().__init__(_)


sob.meta.writable(SecurityScheme).properties = [
    (
        'type_',
        sob.properties.Enumerated(
            values=('apiKey', 'http', 'oauth2', 'openIdConnect'),
            name='type',
            required=True
        )
    ),
    ('description', sob.properties.String()),
    (
        'name',
        sob.properties.String(
            required=lambda o: True if o.type_ == 'apiKey' else False
        )
    ),
    (
        'in_',
        sob.properties.Property(
            types=(
                sob.properties.Enumerated(
                    values=('query', 'header', 'cookie'),
                    versions=('openapi>=3.0',)
                ),
                sob.properties.Enumerated(
                    values=('query', 'header'),
                    versions=('openapi<3.0',)
                ),
            ),
            name='in',
            required=lambda o: True if o.type_ == 'apiKey' else False
        )
    ),
    (
        'scheme',
        sob.properties.String(
            required=lambda o: True if o.type_ == 'http' else False,
            versions='openapi>=3.0'
        )
    ),
    ('bearer_format', sob.properties.String(name='bearerFormat')),
    (
        'flows',
        sob.properties.Property(
            types=(OAuthFlows,),
            required=lambda o: True if o.type_ == 'oauth2' else False,
            versions=('openapi>=3.0',)
        )
    ),
    (
        'open_id_connect_url',
        sob.properties.String(
            name='openIdConnectUrl',
            required=lambda o: True if o.type_ == 'openIdConnect' else False
        )
    ),
    (
        'flow',
        sob.properties.String(
            versions='openapi<3.0',
            required=lambda o: True if o.type_ == 'oauth2' else False
        )
    ),
    (
        'authorization_url',
        sob.properties.String(
            name='authorizationUrl',
            versions='openapi<3.0',
            required=lambda o: (
                True
                if o.type_ == 'oauth2' and o.flow in ('implicit', 'accessCode') else
                False
            )
        )
    ),
    (
        'token_url',
        sob.properties.String(
            name='tokenUrl',
            versions='openapi<3.0',
            required=lambda o: (
                True
                if o.type_ == 'oauth2' and o.flow in ('password', 'application', 'accessCode') else
                False
            )
        )
    ),
    (
        'scopes',
        sob.properties.Dictionary(
            value_types=(str,),
            versions='openapi<3.0',
            required=lambda o: True if o.type_ == 'oauth2' else False
        )
    )
]

sob.meta.writable(Schema).properties = [
    ('title', sob.properties.String()),
    ('description', sob.properties.String()),
    ('multiple_of', sob.properties.Number(name='multipleOf')),
    ('maximum', sob.properties.Number()),
    ('exclusive_maximum', sob.properties.Boolean(name='exclusiveMaximum')),
    ('minimum', sob.properties.Number()),
    ('exclusive_minimum', sob.properties.Boolean(name='exclusiveMinimum')),
    ('max_length', sob.properties.Integer(name='maxLength')),
    ('min_length', sob.properties.Integer(name='minLength')),
    ('pattern', sob.properties.String()),
    ('max_items', sob.properties.Integer(name='maxItems')),
    ('min_items', sob.properties.Integer(name='minItems')),
    ('unique_items', sob.properties.Boolean(name='uniqueItems')),
    (
        'items',
        sob.properties.Property(
            types=(
                Reference,
                Schema,
                sob.properties.Array(
                    item_types=(Reference, Schema)
                )
            )
        )
    ),
    ('max_properties', sob.properties.Integer(name='maxProperties')),
    ('min_properties', sob.properties.Integer(name='minProperties')),
    ('properties', sob.properties.Property(types=(Properties,))),
    (
        'additional_properties',
        sob.properties.Property(
            types=(Reference, Schema, bool),
            name='additionalProperties'
        )
    ),
    ('enum', sob.properties.Array()),
    (
        'type_',
        sob.properties.Property(
            types=(
                sob.properties.Array(
                    item_types=(
                        sob.properties.Enumerated(
                            values=(
                                'array',
                                'object',
                                'file',
                                'integer',
                                'number',
                                'string',
                                'boolean'
                            )
                        ),
                    )
                ),
                sob.properties.Enumerated(
                    values=(
                        'array',
                        'object',
                        'file',
                        'integer',
                        'number',
                        'string',
                        'boolean'
                    )
                )
            ),
            name='type'
        )
    ),
    (
        'format_',
        sob.properties.String(
            # values=(
            #     'int32', 'int64',
            #     'float', 'double',
            #     'byte', 'binary', 'date', 'date-time', 'password'
            # ),
            name='format'
        )
    ),
    ('required', sob.properties.Array(item_types=(sob.properties.String(),))),
    ('all_of', sob.properties.Array(item_types=(Reference, Schema), name='allOf')),
    ('any_of', sob.properties.Array(item_types=(Reference, Schema), name='anyOf')),
    ('one_of', sob.properties.Array(item_types=(Reference, Schema), name='oneOf')),
    ('is_not', sob.properties.Property(types=(Reference, Schema), name='isNot')),
    ('definitions', sob.properties.Property()),
    ('default', sob.properties.Property()),
    ('required', sob.properties.Array(item_types=(str,))),
    ('default', sob.properties.Property()),
    (
        'discriminator',
        sob.properties.Property(
            types=(
                sob.properties.Property(
                    types=(Discriminator,),
                    versions=('openapi>=3.0',),
                ),
                sob.properties.Property(
                    types=(str,),
                    versions=('openapi<3.0',)
                )
            )
        )
    ),
    ('read_only', sob.properties.Boolean()),
    ('write_only', sob.properties.Boolean(name='writeOnly', versions=('openapi>=3.0',))),
    ('xml', sob.properties.Property(types=(XML,))),
    ('external_docs', sob.properties.Property(types=(ExternalDocumentation,))),
    ('example', sob.properties.Property()),
    ('deprecated', sob.properties.Boolean(versions=('openapi>=3.0',))),
    ('links', sob.properties.Array(item_types=(Link,))),
    ('nullable', sob.properties.Boolean(versions=('openapi>=3.0',)))
]


def _schema_after_validate(o):
    # type: (Schema) -> Schema
    if o.format_ in (
        'int32', 'int64',  # type_ == 'integer'
        'float', 'double',  # type_ == 'number'
        'byte', 'binary', 'date', 'date-time', 'password',  # type_ == 'string'
    ):
        if o.type_ == 'integer' and (
            o.format_ not in ('int32', 'int64', None)
        ):
            qn = sob.utilities.qualified_name(type(o))
            raise sob.errors.ValidationError(
                '"%s" in not a valid value for `%s.format_` in this circumstance. ' % (o.format_, qn) +
                '`%s.format_` may be "int32" or "int64" when ' % qn +
                '`%s.type_` is "integer".' % (qn, )
            )
        elif o.type_ == 'number' and (
            o.format_ not in ('float', 'double', None)
        ):
            qn = sob.utilities.qualified_name(type(o))
            raise sob.errors.ValidationError(
                '"%s" in not a valid value for `%s.format_` in this circumstance. ' % (o.format_, qn) +
                '`%s.format_` may be "float" or "double" when ' % qn +
                '`%s.type_` is "number".' % (qn, )
            )
        elif o.type_ == 'string' and (
            o.format_ not in ('byte', 'binary', 'date', 'date-time', 'password', None)
        ):
            qn = sob.utilities.qualified_name(type(o))
            raise sob.errors.ValidationError(
                '"%s" in not a valid value for `%s.format_` in this circumstance. ' % (o.format_, qn) +
                '`%s.format_` may be "byte", "binary", "date", "date-time" or "password" when ' % qn +
                '`%s.type_` is "string".' % (qn, )
            )
    return o


sob.hooks.writable(Schema).after_validate = _schema_after_validate


class Schemas(Dictionary):

    pass


sob.meta.writable(Schemas).value_types = (Reference, Schema)


class Headers(Dictionary):

    pass


sob.meta.writable(Headers).value_types = (Reference, Header)


class Parameters(Dictionary):

    pass


sob.meta.writable(Parameters).value_types = (Reference, Parameter)


class Examples(Dictionary):

    pass


sob.meta.writable(Examples).value_types = (Reference, Example)


class RequestBodies(Dictionary):

    pass


sob.meta.writable(RequestBodies).value_types = (Reference, RequestBody)


class SecuritySchemes(Dictionary):

    pass


sob.meta.writable(SecuritySchemes).value_types = (Reference, SecurityScheme)


class Links(Dictionary):

    pass


sob.meta.writable(Links).value_types = (Reference, Link_)


class Components(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        schemas=None,  # type: Optional[Schemas]
        responses=None,  # type: Optional[Responses]
        parameters=None,  # type: Optional[Parameters]
        examples=None,  # type: Optional[Examples]
        request_bodies=None,  # type: Optional[RequestBodies]
        headers=None,  # type: Optional[Headers]
        security_schemes=None,  # type: Optional[SecuritySchemes]
        links=None,  # type: Optional[Links]
        callbacks=None,  # type: Union[typing.Mapping[str, Union[typing.Dict[str, PathItem], Reference]]]
    ):
        self.schemas = schemas
        self.responses = responses
        self.parameters = parameters
        self.examples = examples
        self.request_bodies = request_bodies
        self.headers = headers
        self.security_schemes = security_schemes
        self.links = links
        self.callbacks = callbacks
        super().__init__(_)


sob.meta.writable(Components).properties = [
    ('schemas', sob.properties.Property(types=(Schemas,))),
    ('responses', sob.properties.Property(types=(Responses,))),
    (
        'parameters',
        sob.properties.Property(types=(Parameters,))
    ),
    ('examples', sob.properties.Property(types=(Examples,))),
    ('request_bodies', sob.properties.Property(types=(RequestBodies,), name='requestBodies')),
    ('headers', sob.properties.Property(types=(Headers,))),
    ('security_schemes', sob.properties.Property(types=(SecuritySchemes,), name='securitySchemes')),
    ('links', sob.properties.Property(types=(Links,))),
    (
        'callbacks',
        sob.properties.Property(
            types=(Callbacks,)
        )
    ),
]


class Definitions(Dictionary):

    pass


sob.meta.writable(Definitions).value_types = (Schema,)


class Paths(Dictionary):

    pass


sob.meta.writable(Paths).value_types = (PathItem,)


class OpenAPI(Object):

    def __init__(
        self,
        _=None,  # type: Optional[typing.Mapping]
        openapi=None,  # type: Optional[str]
        info=None,  # type: Optional[Info]
        host=None,  # type: Optional[str]
        servers=None,  # type: Optional[Sequence[str]]
        base_path=None,  # type: Optional[str]
        schemes=None,  # type: Optional[Sequence[str]]
        tags=None,  # type: Optional[Sequence[Tag]]
        paths=None,  # type: Optional[typing.Mapping[str, PathItem]]
        components=None,  # type: Optional[Components]
        consumes=None,  # type: Any
        # 2.0 compatibility
        swagger=None,  # type: Optional[str]
        definitions=None,  # type: Optional[typing.Mapping[str, Object]]
        security_definitions=None,  # type: Optional[typing.Mapping[str, Union[Object, str]]]
        produces=None,  # type: Optional[typing.Collection[str]]
        external_docs=None,  # type: Optional[ExternalDocumentation]
        parameters=None,  # type: Optional[Dict[str, Parameter]]
        responses=None,  # type: Optional[Dict[str, Response]]
        security=None,  # type: Optional[Dict[str, typing.Sequence[str]]]
    ):
        # type: (...) -> None
        self.openapi = openapi
        self.info = info
        self.host = host
        self.servers = servers
        self.base_path = base_path
        self.schemes = schemes
        self.tags = tags
        self.paths = paths
        self.components = components
        self.consumes = consumes
        # 2.0 compatibility
        self.swagger = swagger
        self.definitions = definitions
        self.security_definitions = security_definitions
        self.produces = produces
        self.external_docs = external_docs
        self.parameters = parameters
        self.responses = responses
        self.security = security
        super().__init__(_)
        version = self.openapi or self.swagger
        if version is not None:
            sob.meta.version(self, 'openapi', version)


sob.meta.writable(OpenAPI).properties = [
    ('openapi', sob.properties.String(versions=('openapi>=3.0',), required=True)),
    ('info', sob.properties.Property(types=(Info,), required=True)),
    ('host', sob.properties.String(versions=('openapi<3.0',))),
    ('servers', sob.properties.Array(item_types=(Server,), versions=('openapi>=3.0',))),
    ('base_path', sob.properties.String(name='basePath', versions=('openapi<3.0',))),
    ('schemes', sob.properties.Array(item_types=(str,), versions=('openapi<3.0',))),
    ('tags', sob.properties.Array(item_types=(Tag,))),
    ('paths', sob.properties.Property(types=(Paths,), required=True)),
    ('components', sob.properties.Property(types=(Components,), versions=('openapi>=3.0',))),
    ('consumes', sob.properties.Array(item_types=(str,), versions=('openapi<3.0',))),
    ('swagger', sob.properties.String(versions=('openapi<3.0',), required=True)),
    (
        'definitions',
        sob.properties.Property(
            types=(Definitions,),
            versions=('openapi<3.0',)
        ),
    ),
    (
        'security_definitions',
        sob.properties.Property(
            types=(
                sob.properties.Dictionary(
                    value_types=(SecurityScheme,),
                    versions=('openapi<3.0',)
                ),
                sob.properties.Dictionary(
                    value_types=(Reference, SecurityScheme),
                    versions=('openapi>=3.0',)
                ),
            ),
            name='securityDefinitions',
        )
    ),
    ('produces', sob.properties.Array(item_types=(str,), versions=('openapi<3.0',)),),
    ('external_docs', sob.properties.Property(types=(ExternalDocumentation,), name='externalDocs')),
    (
        'parameters',
        sob.properties.Dictionary(
            value_types=(Parameter,),
            versions=('openapi<3.0',)
        )
    ),
    ('responses', sob.properties.Dictionary(value_types=(Response,), versions=('openapi<3.0',))),
    (
        'security',
        sob.properties.Dictionary(
            value_types=(
                sob.properties.Array(item_types=(str,)),
            ),
            versions=('openapi<3.0',)
        )
    ),
]



from setuptools import setup, find_packages

setup(
    name='oapi',
    version="0.0.63",
    description=(
        'An SDK for parsing OpenAPI (Swagger) 2.0 - 3.0 specifications'
    ),
    url='https://bitbucket.com/davebelais/oapi.git',
    author='David Belais',
    author_email='david@belais.me',
    license='MIT',
    python_requires='>=3.6, <4',
    keywords='openapi swagger json rest',
    packages=[
        'oapi',
        'oapi.oas'
    ],
    install_requires=[
        "pyyaml>=5.3",
        "iso8601>=0.1.12",
        "sob>=0.2.13",
        "jsonpointer>=2.0"
    ],
    extras_require={
        "dev": [
            "pytest>=5.4.1",
            "tox>=3.14.5",
            "flake8>=3.7.9"
        ],
        "test": [
            "pytest>=5.4.1",
            "tox>=3.14.5",
            "flake8>=3.7.9"
        ]
    }
)



import os
import re
import collections

from collections import OrderedDict
from io import IOBase
from urllib.parse import urlparse, urljoin
from urllib.request import urlopen

from sob import __name__ as _sob_module_name
from sob import meta
from sob import properties
from sob.utilities import (
    get_source, property_name, properties_values, class_name,
    calling_function_qualified_name
)
from sob.utilities.string import split_long_comment_line
from sob.model import (
    Model as ModelBase, Object, Dictionary, Array,
    from_meta as model_from_meta
)
from sob.utilities import qualified_name, url_relative_to
from sob.properties import Property
from sob.utilities.types import Null
from .oas.references import Resolver
from .oas.model import OpenAPI, Schema, Reference, Parameter, Operation
from typing import Iterable, Sequence, Optional, Union, List

_META_MODULE_QUALIFIED_NAME = qualified_name(meta)
_META_PROPERTIES_QAULIFIED_NAME = qualified_name(meta.Properties)
_META_PROPERTIES_QAULIFIED_NAME_LENGTH = len(_META_PROPERTIES_QAULIFIED_NAME)
_DOC_POINTER_RE = re.compile(
    (
        # Pointer
        r'^(.*?)'
        # Pointer stops at a double-return or end-of-string
        r'(?:\r?\n\s*\r?\n|$)'
    ),
    re.DOTALL
)
_SPACES_RE = re.compile(r'[\s\n]')
_LINE_LENGTH = 79  # type: int

# region Functions


def schema_defines_array(schema):
    # type: (Schema) -> bool
    """
    Schemas of the type `array` translate to an `Array`. Incorrectly
    implemented schemas may also neglect this, however use of the attribute
    `items` is only valid in the context of an array--and therefore indicate
    the schema also defines an instance of `Array`.
    """
    return schema.type_ == 'array' or schema.items


def schema_defines_object(schema):
    # type: (Schema) -> bool
    """
    If properties are defined for a schema, it will translate to an instance of
    `Object`.
    """
    return (
        schema.properties and
        schema.type_ in ('object', None) and
        (not schema.additional_properties)
    )


def schema_defines_dictionary(schema):
    # type: (Schema) -> bool
    """
    If properties are not defined for a schema, or unspecified attributes are
    allowed--the schema will translate to an instance of `Dictionary`.
    """
    return schema.type_ == 'object' and (
        schema.additional_properties or (not schema.properties)
    )


def schema_defines_model(schema):
    # type: (Schema) -> bool
    return (
        schema_defines_array(schema) or
        schema_defines_object(schema) or
        schema_defines_dictionary(schema)
    )


def operation_defines_model(operation):
    # type: (Operation) -> bool
    """
    Determine if an operation requires a model (if it describes a form)
    """
    if operation.parameters:
        for parameter in operation.parameters:
            if parameter.in_ == 'form':
                return True
    return False


# endregion


# region Private Classes


class _RecursiveReferencePlaceholder:

    def __init__(self, relative_url_pointer):
        # type: (str) -> None
        self.relative_url_pointer = relative_url_pointer

    def __eq__(self, other):
        # type: (_RecursiveReferencePlaceholder) -> bool
        return (
            self.relative_url_pointer == other.relative_url_pointer
        )


class _Modeler:
    """
    This class parses an OpenAPI schema and produces a data model based on the
    `sob` library.
    """

    def __init__(self, root):
        # type: (Union[IOBase, str], str, Callable) -> None
        if not isinstance(root, OpenAPI):
            root = OpenAPI(root)
        # This ensures all elements have URLs and JSON pointers
        meta.url(root, meta.url(root))
        meta.pointer(root, meta.pointer(root))
        # Private Properties
        self._traversed_relative_urls_pointers = set()
        self._relative_urls_pointers_class_names = {}
        self._class_names_relative_urls_pointers = {}
        self._relative_urls_pointers_meta = {}
        self._class_names_meta = {}
        self._relative_urls_pointers_models = {}
        self._class_names_models = {}
        # Public properties
        self.root = root
        self.resolver = Resolver(self.root)
        self.major_version = int(
            (self.root.swagger or self.root.openapi).split('.')[0].strip()
        )
        self.references = OrderedDict()
        self.pointers_schemas = OrderedDict()
        self.names_models = OrderedDict()
        self.names_schemas = {}
        self.pointers_models = OrderedDict()

    def get_relative_url_pointer_model(self, relative_url_pointer):
        # type: (str) -> Module
        return self._relative_urls_pointers_models[relative_url_pointer]

    def set_relative_url_pointer_class_name(
        self, relative_url_pointer, class_name_
    ):
        # type: (str, str) -> str
        # Check to see if this pointer already has a class name, and use that
        # if it does
        if relative_url_pointer in self._relative_urls_pointers_class_names:
            class_name_ = self._relative_urls_pointers_class_names[
                relative_url_pointer
            ]
        else:
            self._relative_urls_pointers_class_names[
                relative_url_pointer
            ] = class_name_
        self._class_names_relative_urls_pointers[
            class_name_
        ] = relative_url_pointer
        return class_name_

    def relative_url_pointer_class_name_exists(self, relative_url_pointer):
        # type: (str) -> bool
        return relative_url_pointer in self._relative_urls_pointers_class_names

    def relative_url_pointer_model_exists(self, relative_url_pointer):
        # type: (str) -> bool
        return relative_url_pointer in self._relative_urls_pointers_models

    def class_name_relative_urls_pointer_exists(self, class_name_):
        # type: (str) -> bool
        return class_name_ in self._class_names_relative_urls_pointers

    def get_relative_url_pointer_class_name(self, relative_url_pointer):
        # type: (str) -> str
        return self._relative_urls_pointers_class_names[relative_url_pointer]

    def get_class_name_relative_url_pointer(self, class_name_):
        # type: (str) -> str
        return self._class_names_relative_urls_pointers[class_name_]

    def property_any_of(self, property_, schemas):
        # type: (Property, Iterable[Schema]) -> None
        if schemas:
            if property_.types is None:
                property_.types = []
            for schema in schemas:
                child_property_ = self.get_property(schema)
                if child_property_ is not None:
                    property_.types.append(child_property_)

    def property_all_of(self, property_, schemas):
        # type: (Iterable[Schema]) -> None
        """
        TODO
        """
        if schemas:
            if property_.types is None:
                property_.types = []
            for schema in schemas:
                child_property_ = self.get_property(schema)
                if child_property_ is not None:
                    property_.types.append(child_property_)

    def get_property(
        self,
        schema,  # type: Union[Schema, Reference]
        name=None,  # type: Optional[str]
        required=None  # type: Optional[bool]
    ):
        # type: (...) -> properties.Property
        is_referenced = isinstance(schema, Reference)
        schema = self._resolve(schema)
        if (
            (schema.any_of is not None) or
            (schema.one_of is not None) or
            (schema.all_of is not None)
        ):
            property_ = properties.Property()
            self.property_any_of(property_, schema.any_of)
            self.property_any_of(property_, schema.one_of)
            self.property_all_of(property_, schema.all_of)
        elif schema_defines_model(schema):
            property_ = properties.Property()
            property_types = self.get_model(
                schema,
                is_property=(not is_referenced)
            )
            if property_types is not None:
                property_.types = (property_types,)
        elif schema.type_ == 'number':
            property_ = properties.Number()
        elif schema.type_ == 'integer':
            property_ = properties.Integer()
        elif schema.type_ == 'string':
            if schema.format_ == 'date-time':
                property_ = properties.DateTime()
            elif schema.format_ == 'date':
                property_ = properties.Date()
            elif schema.format_ == 'byte':
                property_ = properties.Bytes()
            else:
                property_ = properties.String()
        elif schema.type_ == 'boolean':
            property_ = properties.Boolean()
        elif schema.type_ == 'file':
            property_ = properties.Bytes()
        else:
            raise ValueError(
                'No schema "type" found:\n' + repr(schema)
            )
        if schema.enum:
            property_ = properties.Enumerated(
                values=tuple(schema.enum),
                types=(property_,)
            )
        if name is not None:
            property_.name = name
        if (
            schema.nullable or
            # Swagger/OpenAPI versions prior to 3.0 do not support `nullable`,
            # so it must be assumed that null values are acceptable for
            # required attributes
            ((self.major_version < 3) and (required is True))
        ):
            if schema.nullable is not False:
                name, required, versions = (
                    property_.name, property_.required, property_.versions
                )
                property_.name = (
                    property_.required
                ) = property_.versions = None
                property_ = Property(
                    types=(property_, Null),
                    name=name,
                    required=required,
                    versions=versions
                )
        if required is not None:
            property_.required = required
        return property_

    def get_relative_url(self, url):
        # type: (str) -> str
        parse_result = urlparse(url)
        # Determine if the URL is absolute or relative
        if parse_result.netloc or parse_result.scheme == 'file':
            # Only include the relative URL if it is not the root document
            if url == self.resolver.url:
                relative_url = ''
            else:
                relative_url = url_relative_to(url, self.resolver.url)
        else:
            relative_url = url
        return relative_url

    def get_definition_relative_url_and_pointer(self, definition):
        # type: (Union[Schema, Operation, Parameter]) -> str
        """
        Given a schema/operation/parameter definition, return a relative path
        in relation to the root document and the pointer
        """
        url = meta.url(definition)
        pointer = meta.pointer(definition)
        return self.get_relative_url(url), pointer

    def get_definition_relative_url_pointer(self, definition):
        # type: (Union[Schema, Operation, Parameter]) -> str
        """
        Given a schema/operation/parameter definition, return a relative path +
        pointer in relation to the root document
        """
        relative_url, pointer = self.get_definition_relative_url_and_pointer(
            definition
        )
        return relative_url + pointer

    def set_relative_url_pointer_model(self, relative_url_pointer, model):
        # type: (str, Optional[sob.model.Model]) -> None
        self._relative_urls_pointers_models[relative_url_pointer] = model

    def set_model_class_name(self, model, class_name_=None):
        # type: (Optional[sob.model.Model], Optional[str]) -> None
        if class_name_ is None:
            class_name_ = model.__name__
        self._class_names_models[class_name_] = model

    def get_model(self, definition, is_property=False):
        # type: (Union[Schema, Operation, Parameter], bool) -> ModelBase
        relative_url_pointer = self.get_definition_relative_url_pointer(
            definition
        )
        # If this model has already been generated--use the cached model
        if not self.relative_url_pointer_model_exists(relative_url_pointer):
            # Setting this to prevents recursion errors
            self.set_relative_url_pointer_model(
                relative_url_pointer,
                None
            )
            if isinstance(definition, Schema):
                model = self.get_schema_model(
                    definition,
                    relative_url_pointer=relative_url_pointer,
                    is_property=is_property
                )
            elif isinstance(definition, Operation):
                model = self.get_operation_model(
                    definition,
                    relative_url_pointer=relative_url_pointer,
                    is_property=is_property
                )
            elif isinstance(definition, Parameter):
                model = self.get_parameter_model(
                    definition,
                    relative_url_pointer=relative_url_pointer,
                    is_property=is_property
                )
            else:
                raise TypeError(
                    'The model definition must be a `%s`, `%s` or `%s`, '
                    'not:\n%s'
                    % (
                        qualified_name(Schema),
                        qualified_name(Operation),
                        qualified_name(Parameter),
                        repr(definition)
                    )
                )
            self.set_relative_url_pointer_model(
                relative_url_pointer,
                model
            )
            self.set_model_class_name(model)
        return self.get_relative_url_pointer_model(relative_url_pointer)

    @property
    def models(self):
        # type: (...) -> typing.Iterator[Union[Schema, Operation, Parameter]]
        models_names = {}
        for definition in self.model_definitions:
            model = self.get_model(definition)
            if model.__name__ in models_names:
                existing_model_meta = meta.read(models_names[model.__name__])
                new_model_meta = meta.read(model)
                # Ensure this is just a repeat use of the same model, and not
                # a different model of the same name
                if existing_model_meta != new_model_meta:
                    raise RuntimeError(
                        'An attempt was made to define a model using an '
                        'existing model name (`%s`)"\n\n'
                        'Existing Module Metadata:\n\n%s\n\n'
                        'New Module Metadata:\n\n%s' % (
                            model.__name__,
                            repr(existing_model_meta),
                            repr(new_model_meta)
                        )
                    )
            elif model not in (Array, Dictionary):
                models_names[model.__name__] = model
                yield model

    def get_operation_model(self, operation):
        # type: (Operation) -> type
        name = self.get_operation_class_name(operation)
        # TODO

    def get_parameter_model(self, parameter):
        # type: (Parameter) -> type
        return self.get_schema_model(
            parameter.schema,
            name=self.get_parameter_class_name(parameter)
        )

    def get_schema_model(
        self, schema, name=None, relative_url_pointer=None, is_property=False
    ):
        # type: (Schema, Optional[str], Optional[str], bool) -> type
        if name is None:
            name = self.get_schema_class_name(
                schema,
                is_property=is_property
            )
        if relative_url_pointer is None:
            relative_url_pointer = self.get_definition_relative_url_pointer(
                schema
            )
        if schema_defines_array(schema):
            return self.get_schema_array(
                schema,
                name=name,
                relative_url_pointer=relative_url_pointer
            )
        elif schema_defines_object(schema):
            return self.get_schema_object(
                schema,
                name=name,
                relative_url_pointer=relative_url_pointer
            )
        elif schema_defines_dictionary(schema):
            return self.get_schema_dictionary(
                schema,
                name=name,
                relative_url_pointer=relative_url_pointer
            )
        else:
            return None

    def get_schema_array(self, schema, name=None, relative_url_pointer=None):
        # type: (Schema, Optional[str], Optional[str]) -> type
        # Get all applicable items schemas
        items_schemas = self._resolve(schema.items)
        if isinstance(items_schemas, (collections.Sequence, collections.Set)):
            items_schemas = tuple(
                self._resolve(items_schema) for items_schema in items_schemas
            )
        elif items_schemas is not None:
            items_schemas = (items_schemas,)
        item_types = None
        # If item types are defined--create a class, otherwise--use the base
        # class
        if items_schemas:
            item_types = []
            for items_schema in items_schemas:
                if schema_defines_model(items_schema):
                    items_model = self.get_model(items_schema)
                else:
                    items_model = self.get_property(items_schema)
                # Checking to make sure the model is not `None` is
                # necessary because is some fringe cases (involving cyclic
                # references) a model cannot be retrieved at this point
                if items_model is not None:
                    item_types.append(items_model)
        if item_types:
            array_meta = meta.Array(item_types=item_types)
            # If the array allows a single type--base the array name on that
            # type in order to allow re-use
            if (
                len(item_types) == 1 and isinstance(item_types[0], type) and
                issubclass(item_types[0], ModelBase)
            ):
                name = item_types[0].__name__
            elif name is None:
                name = self.get_schema_class_name(schema)
            name += 'Array'
            if name in self._class_names_models:
                return self._class_names_models[name]
            else:
                if relative_url_pointer is None:
                    relative_url_pointer = (
                        self.get_definition_relative_url_pointer(schema)
                    )
                name = self.set_relative_url_pointer_class_name(
                    relative_url_pointer, name
                )
                self._class_names_meta[name] = array_meta
                array_class = model_from_meta(
                    name,
                    array_meta,
                    docstring=self.get_docstring(schema, relative_url_pointer),
                    module='__main__'
                )
        else:
            array_class = Array
        return array_class

    def get_schema_dictionary(
        self, schema, name=None, relative_url_pointer=None
    ):
        # type: (Schema, Optional[str], Optional[str]) -> type
        # Get the value type schema, if applicable
        if isinstance(schema.additional_properties, (Schema, Reference)):
            if name is None:
                name = self.get_schema_class_name(schema)
            if relative_url_pointer is None:
                relative_url_pointer = (
                    self.get_definition_relative_url_pointer(schema)
                )
            name += 'Dictionary'
            dictionary_meta = meta.Dictionary()
            value_type = self.get_model(
                self._resolve(
                    schema.additional_properties
                )
            )
            dictionary_meta.value_types = (value_type,)
            name = self.set_relative_url_pointer_class_name(
                relative_url_pointer, name
            )
            self._class_names_meta[name] = dictionary_meta
            dictionary_class = model_from_meta(
                name,
                dictionary_meta,
                docstring=self.get_docstring(schema, relative_url_pointer),
                module='__main__'
            )
        else:
            dictionary_class = Dictionary
        return dictionary_class

    def get_docstring(self, schema, relative_url_pointer):
        # type: (Schema, str) -> type
        if relative_url_pointer is None:
            relative_url_pointer = self.get_definition_relative_url_pointer(
                schema
            )

        if len(relative_url_pointer) > 116 and relative_url_pointer[0] != '#':
            pointer_split = relative_url_pointer.split('#')
            docstring_lines = [
                pointer_split[0] + '\n' +
                '#' + '#'.join(pointer_split[1:])
            ]
        else:
            docstring_lines = [relative_url_pointer]
        if schema and schema.description:
            docstring_lines.append(schema.description)
        return '\n\n'.join(docstring_lines)

    def get_schema_object(self, schema, name=None, relative_url_pointer=None):
        # type: (Schema, Optional[str], Optional[str]) -> type
        if name is None:
            name = self.get_schema_class_name(schema)
        if relative_url_pointer is None:
            relative_url_pointer = self.get_definition_relative_url_pointer(
                schema
            )
        object_meta = meta.Object()
        for name_, property_ in schema.properties.items():
            property_name_ = property_name(name_)
            # Prevent property names from conflicting with the dependency
            # module namespace
            if property_name_ == _sob_module_name:
                property_name_ = _sob_module_name + '_'
            object_meta.properties[property_name_] = self.get_property(
                property_,
                name=None if property_name_ == name_ else name_,
                required=(
                    True if (
                        schema.required and (
                            name_ in schema.required
                        )
                    ) else False
                )
            )
        name = self.set_relative_url_pointer_class_name(
            relative_url_pointer, name
        )
        self._class_names_meta[name] = object_meta
        return model_from_meta(
            name,
            object_meta,
            docstring=self.get_docstring(schema, relative_url_pointer),
            module='__main__'
        )

    def get_operation_class_name(self, operation):
        # type: (Operation) -> str
        """
        Derive a model's class name from an `Operation` object. This only
        applies when one or more parameters are
        "in": "form".
        """
        relative_url, pointer = self.get_definition_relative_url_and_pointer(
            operation
        )
        return self.get_relative_url_and_pointer_class_name(
            relative_url, pointer, is_property=True
        )

    def get_parameter_class_name(self, parameter):
        # type: (Parameter) -> str
        """
        Derive a model's class name from a `Parameter` object
        """
        relative_url, pointer = self.get_definition_relative_url_and_pointer(
            parameter
        )
        return self.get_relative_url_and_pointer_class_name(
            relative_url, pointer, is_property=True
        )

    def get_schema_class_name(self, schema, is_property=False):
        # type: (Schema, bool) -> str
        """
        Derive a model's class name from a `Schema` object
        """
        relative_url, pointer = self.get_definition_relative_url_and_pointer(
            schema
        )
        return self.get_relative_url_and_pointer_class_name(
            relative_url, pointer, is_property=is_property
        )

    def get_relative_url_and_pointer_class_name(
        self, relative_url, pointer, is_property=False
    ):
        # type: (str, str, Optional[bool]) -> str
        relative_url_pointer = relative_url + pointer
        if not self.relative_url_pointer_class_name_exists(relative_url_pointer):
            parts = [
                part.replace('~1', '/') for part in
                pointer.lstrip('#/').split('/')
            ]
            length = len(parts)
            start_index = 0
            first_part = parts[0].lower()
            if length > 1:
                if first_part in ('definitions', 'paths'):
                    start_index += 1
                elif first_part == 'components':
                    start_index += 1
                    if length > 2 and parts[1] == 'schemas':
                        start_index += 1

            def get_class_name(pointer_start_index):
                # type: (int) -> str
                relevant_parts = list(parts[pointer_start_index:])
                if relative_url:
                    relevant_parts.insert(0, relative_url)
                if is_property and len(relevant_parts) > 1:
                    relevant_parts.pop(-2)
                return class_name(
                    '/'.join(relevant_parts)
                )

            relative_url_pointer_class_name = get_class_name(start_index)
            while start_index and (
                    self.class_name_relative_urls_pointer_exists(relative_url_pointer_class_name) and
                    self.get_class_name_relative_url_pointer(
                    relative_url_pointer_class_name
                ) != relative_url_pointer
            ):
                start_index -= 1
                relative_url_pointer_class_name = get_class_name(start_index)
            # Cache the result
            self.set_relative_url_pointer_class_name(
                relative_url_pointer, relative_url_pointer_class_name
            )
        return self.get_relative_url_pointer_class_name(relative_url_pointer)

    @property
    def model_definitions(self):
        # type: (...) -> Tuple[str, str, ModelBase]
        """
        This property returns all objects defining a model, namely instances of
        `Schema`, `Parameter`, and `Operation`
        """
        self._traversed_relative_urls_pointers = set()
        for definition in tuple(
            self._traverse_model_definitions(
                self.root, (OpenAPI,)
            )
        ):
            yield self._resolve(definition)

    def _resolve(self, model_instance, types=None):
        # type: (ModelBase, Optional[Sequence[type, Property]]) -> ModelBase
        """
        If `model_instance` is a reference, get the referenced object
        """
        if isinstance(model_instance, Reference):
            if types is None:
                types = (Schema,)
            url = meta.url(model_instance)
            pointer = urljoin(
                meta.pointer(model_instance),
                model_instance.ref
            )
            while isinstance(model_instance, Reference):
                resolved_model_instance = self.resolver.get_document(
                    url
                ).resolve(
                    pointer,
                    types
                )
                if resolved_model_instance is model_instance:
                    raise RuntimeError(
                        '`Reference` instance is self-referential: ' +
                        pointer
                    )
                else:
                    model_instance = resolved_model_instance
            if isinstance(model_instance, Reference):
                raise RuntimeError(pointer)
        return model_instance

    def _add_traversed_definition(
        self,
        definition  # type: ModelBase
    ):
        # type: (...) -> None
        self._traversed_relative_urls_pointers.add(
            self.get_definition_relative_url_pointer(definition)
        )

    def _is_traversed_definition(
        self,
        definition  # type: ModelBase
    ):
        # type: (...) -> bool
        relative_url_pointer = self.get_definition_relative_url_pointer(
            definition
        )
        return relative_url_pointer in self._traversed_relative_urls_pointers

    def _traverse_model_definitions(
        self,
        model_instance,  # type: ModelBase
        types=None  # type: Optional[Union[type, properties.Property]]
    ):
        # type: (...) -> typing.Iterator[Union[Schema, Operation, Parameter]]
        assert isinstance(model_instance, ModelBase)
        model_instance = self._resolve(model_instance, types)
        if not self._is_traversed_definition(model_instance):
            self._add_traversed_definition(model_instance)
            # If this is a schema/parameter/operation defining a model--include
            # it in the returned results
            if (
                isinstance(model_instance, Schema) and
                schema_defines_model(model_instance)
            ):
                yield model_instance
            elif isinstance(model_instance, Parameter):
                if model_instance.schema is not None:
                    schema = self._resolve(model_instance.schema)
                    if schema_defines_model(schema) and not (
                        self._is_traversed_definition(schema)
                    ):
                        yield schema
            elif isinstance(
                model_instance, Operation
            ) and operation_defines_model(model_instance):
                yield model_instance
            meta_ = meta.read(model_instance)
            # Recursively find other definitions
            if isinstance(model_instance, Dictionary):
                for key, value in model_instance.items():
                    if isinstance(value, ModelBase):
                        for definition in self._traverse_model_definitions(
                            value, meta_.value_types
                        ):
                            yield definition
            elif isinstance(model_instance, Array):
                for item in model_instance:
                    if isinstance(item, ModelBase):
                        for definition in self._traverse_model_definitions(
                            item, meta_.item_types
                        ):
                            yield definition
            elif isinstance(model_instance, Object):
                for name, property_ in meta_.properties.items():
                    value = getattr(model_instance, name)
                    if isinstance(value, ModelBase):
                        for definition in self._traverse_model_definitions(
                            value, (property_,)
                        ):
                            yield definition
            else:
                raise TypeError(model_instance)

    def represent_model_meta(self, class_name_):
        # type: (str) -> str
        meta_ = self._class_names_meta[class_name_]
        relative_url_pointer = self.get_class_name_relative_url_pointer(
            class_name_
        )
        lines = list()
        lines.append(
            split_long_comment_line(
                '# ' + relative_url_pointer
            )
        )
        for property_name_, value in properties_values(meta_):
            if value is not None:
                value = repr(value)
                if value[:_META_PROPERTIES_QAULIFIED_NAME_LENGTH + 1] == (
                    _META_PROPERTIES_QAULIFIED_NAME + '('
                ):
                    value = value[
                        _META_PROPERTIES_QAULIFIED_NAME_LENGTH + 1:
                        -1
                    ]
                lines.append(
                    '%s.writable(\n    %s%s\n).%s = %s' % (
                        _META_MODULE_QUALIFIED_NAME,
                        class_name_,
                        (
                            '  # noqa'
                            if len(class_name_) + 4 > _LINE_LENGTH else
                            ''
                        ),
                        property_name_,
                        '\n'.join(
                            line + (
                                '  # noqa'
                                if len(line) > _LINE_LENGTH else ''
                            )
                            for line in value.split('\n')
                        )
                    )
                )
        return '\n'.join(lines) + '\n'

    @staticmethod
    def represent_model(model):
        # type: (ModelBase) -> Tuple[str, str]
        return get_source(model).split('\n\n\n')

    @property
    def module_definition(self) -> str:
        class_names_sources = OrderedDict()
        classes: List[str] = []
        imports: List[str] = []
        for model_class in self.models:
            class_name_ = model_class.__name__
            imports_str, source = self.represent_model(model_class)
            if class_name_ in class_names_sources:
                raise RuntimeError(
                    'The class name `%s` occured twice:\n\n%s\n\n%s' % (
                        class_name_,
                        class_names_sources[class_name_],
                        source
                    )
                )
            class_names_sources[class_name_] = source
            # Only include each import once
            line: str
            for line in imports_str.split('\n'):
                if line not in imports:
                    imports.append(line)
            classes.append(source)
        return '\n'.join(
            sorted(
                imports,
                key=lambda sort_line: 1 if sort_line == 'import sob' else 0
            ) + ['\n'] +
            classes + [
                self.represent_model_meta(class_name_)
                for class_name_ in class_names_sources.keys()
            ]
        )


class _ModuleParser:

    def __init__(self, path=None):
        # type: (Optional[str]) -> None
        self._relative_urls_pointers_class_names = None  # type: Optional[dict]
        self.namespace = {}
        if path is not None:
            self.open(path)

    def open(self, path=None):
        # type: (Optional[str]) -> None
        path = os.path.abspath(path)
        current_directory = os.path.abspath(os.curdir)
        os.chdir(os.path.dirname(path))
        self.namespace['__file__'] = path
        with open(path) as module_io:
            exec(module_io.read(), self.namespace)
        os.chdir(current_directory)

    @property
    def models(self):
        # type: (...) -> Iterable[type]
        for name, value in self.namespace.items():
            if (
                name[0] != '_' and isinstance(value, type) and
                issubclass(value, ModelBase)
            ):
                yield value

    @staticmethod
    def _get_model_relative_url_pointer_and_class_name(model):
        # type: (type) -> str
        match = _DOC_POINTER_RE.search(
            model.__doc__
        )
        relative_url_pointer = None
        if match:
            groups = match.groups()
            if groups:
                relative_url_pointer = _SPACES_RE.sub('', groups[0])
        return relative_url_pointer, model.__name__

    @property
    def relative_urls_pointers_class_names(self):
        # type: (...) -> Iterator[str, str]
        for model in self.models:
            relative_url_pointer, name = (
                self._get_model_relative_url_pointer_and_class_name(model)
            )
            if relative_url_pointer is not None:
                yield relative_url_pointer, name


# endregion


# region Public Classes


class Module:
    """
    This class parses an Open API document and outputs a module defining
    classes to represent each schema defined in the Open API document as a
    subclass of `sob.model.Module`.

    Initialization Parameters:

    - data (str|http.client.HTTPResponse|io.IOBase|oapi.oas.model.OpenAPI):

      The input data can be a URL, file-path, an HTTP response
      (`http.client.HTTPResponse`), a file object, or an
      instance of `oapi.oas.model.OpenAPI`.
    """

    def __init__(self, data):
        # type: (Union[str, IOBase, OpenAPI]) -> None
        self._modeler = None  # type: _Modeler
        self._parser = _ModuleParser()
        if isinstance(data, str):
            if os.path.exists(data):
                self._path_init(data)
            else:
                self._url_init(data)
        elif isinstance(data, IOBase):
            self._io_init(data)
        elif isinstance(data, OpenAPI):
            self._model_init(data)
        else:
            raise TypeError(
                '`%s` requires an instance of `str`, `%s`, or `%s` for the '
                '`data` parameter--not %s' % (
                    calling_function_qualified_name(),
                    qualified_name(IOBase),
                    qualified_name(OpenAPI),
                    repr(data)
                )
            )

    def _path_init(self, path):
        # type: (str) -> None
        with open(path, 'r') as model_io:
            self._io_init(model_io)

    def _url_init(self, url):
        # type: (str) -> None
        with urlopen(url) as model_io:
            self._io_init(model_io)

    def _io_init(self, model_io):
        # type: (IOBase) -> None
        self._model_init(OpenAPI(model_io))

    def _model_init(self, model):
        # type: (OpenAPI) -> None
        self._modeler = _Modeler(model)

    def __str__(self):
        # type: (...) -> str
        return self._modeler.module_definition

    def _parse_existing_module(self, path):
        # type: (str) -> None
        if os.path.exists(path):
            self._parser.open(path)
            for relative_url_pointer, class_name_ in (
                self._parser.relative_urls_pointers_class_names
            ):
                self._modeler.set_relative_url_pointer_class_name(
                    relative_url_pointer, class_name_
                )

    def _get_relative_url_and_pointer(self, url_pointer):
        # type: (str) -> str
        relative_url = ''
        pointer = url_pointer
        if url_pointer[0] != '#':
            if '#' in url_pointer:
                key_split = url_pointer.split('#')
                relative_url = self._modeler.get_relative_url(key_split[0])
                pointer = '#' + '#'.join(key_split[1:])
            else:
                relative_url = url_pointer
                pointer = ''
        return relative_url, pointer

    def model_from_pointer(self, url_pointer):
        # type: (str) -> Module
        """
        Return a model from a pointer
        """
        relative_url, pointer = self._get_relative_url_and_pointer(url_pointer)
        return self._modeler.get_relative_url_pointer_model(
            relative_url + pointer
        )

    def save(self, path):
        # type: (str) -> None

        """
        This method will save the generated module to a given path. If there is
        an existing module at that path--the existing module will be imported
        and the pre-existing class names will be utilized for any schemas
        defined by elements residing at the same location as the documented
        JSON pointer in the pre-existing classes' docstrings.
        """
        # Make sure that any matching, existing classes use the same names
        self._parse_existing_module(path)
        # Save the module
        with open(path, 'w') as model_io:
            model_io.write(str(self))

    def __eq__(self, other):
        # type: (Module) -> bool
        return isinstance(other, self.__class__) and str(self) == str(other)


# For backwards compatibility...
Model = Module

# endregion